import os
import time
import itertools

from collections import defaultdict, Callable
import numpy as np
import sympy
import matplotlib.pyplot as plt
from multiprocessing import Pool
import networkx as nx

from qulacs import Observable, QuantumCircuit, QuantumState
from qulacs.gate import RX, RZ, CNOT, merge
from Qcover.backends import Backend
from Qcover.utils import get_graph_weights
import warnings
warnings.filterwarnings("ignore")


class CircuitByQulacs(Backend):
    """generate a instance of CircuitByQulacs"""

    def __init__(self,
                 research: str = "QAOA",
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByCirq"""
        super(CircuitByQulacs, self).__init__()

        self._p = None
        self._origin_graph = None
        self._is_parallel = False if is_parallel is None else is_parallel
        self._research = research

        self._nodes_weight = None
        self._edges_weight = None
        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []
        self._element_expectation = dict()

    @property
    def element_expectation(self):
        return self._element_expectation

    @staticmethod
    def get_operator(element, qubit_num):
        op = Observable(qubit_num)
        if isinstance(element, int):
            op.add_operator(1.0, "Z " + str(element))
        else:
            op.add_operator(1.0, "Z " + str(element[0]) + " Z " + str(element[1]))

        return op

    def get_QAOA_circuit(self, p, graph, node_to_qubit):  #, params=None
        circ = QuantumCircuit(len(graph))
        # if params is None:
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        for k in range(p):
            for i in graph.nodes:
                u = node_to_qubit[i]
                if k == 0:
                    circ.add_H_gate(u)
                circ.add_RZ_gate(u, 2 * gamma_list[k] * self._nodes_weight[i])

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.add_CNOT_gate(u, v)
                circ.add_RZ_gate(v, 2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]])
                circ.add_CNOT_gate(u, v)

            for nd in graph.nodes:
                u = node_to_qubit[nd]
                circ.add_RX_gate(u, 2 * beta_list[k])

        return circ

    def get_GHZ_circuit(self, p, graph, node_to_qubit):
        circ = QuantumCircuit(len(graph))
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]

        pivot = len(self._origin_graph) // 2
        for k in range(p):
            for nd in graph.nodes:
                if k == 0:
                    circ.add_H_gate(node_to_qubit[nd])

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.add_CNOT_gate(u, v)
                circ.add_RZ_gate(v, 2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]])
                circ.add_CNOT_gate(u, v)

            for nd in graph.nodes:
                if nd != pivot:
                    circ.add_RX_gate(node_to_qubit[nd], 2 * beta_list[k])

        return circ

    def get_expectation(self, element_graph, p=None):  #,
        """
        calculate the expectation of the subgraph
        Args:
            element_graph: tuple of (original node/edge, subgraph)

        Returns:
            expectation of the subgraph
        """
        if self._is_parallel is False:
            p = self._p if p is None else p
            original_e, graph = element_graph
        else:
            p = self._p if len(element_graph) == 1 else element_graph[1]
            original_e, graph = element_graph[0]

        node_list = list(graph.nodes)
        node_to_qubit = defaultdict(int)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        state = QuantumState(len(graph.nodes))
        state.set_zero_state()

        if self._research == "QAOA":
            circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
        elif self._research == "GHZ":
            pivot = len(self._origin_graph) // 2
            if pivot in graph:
                circ = self.get_GHZ_circuit(p, graph, node_to_qubit)
            else:
                circ = self.get_QAOA_circuit(p, graph, node_to_qubit)

        circ.update_quantum_state(state)

        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            op = self.get_operator(node_to_qubit[original_e], len(node_list))
        else:
            weight = self._edges_weight[original_e]
            op = self.get_operator((node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]), len(node_list))

        exp_res = op.get_expectation_value(state)

        return weight, exp_res

    def expectation_calculation(self, p=None):
        if self._nodes_weight is None or self._edges_weight is None:
            nodes_weight, edges_weight = get_graph_weights(self._origin_graph)
            self._nodes_weight, self._edges_weight = nodes_weight, edges_weight

        self._element_expectation = {}
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        res = 0
        for item in self._element_to_graph.items():
            w_i, exp_i = self.get_expectation(item, p)
            if isinstance(item[0], tuple):
                self._element_expectation[item[0]] = exp_i
            res += w_i * exp_i

        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        args = list(itertools.product(self._element_to_graph.items(), [p]))
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, args))
        # circ_res.append(pool.map(self.get_expectation, list(self._element_to_graph.items()), chunksize=1))

        pool.terminate()  # pool.close()
        pool.join()
        res = 0
        for it in circ_res[0]:
            res += it[0] * it[1]
        # res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def get_result_counts(self, params):
        node_list = list(self._origin_graph.nodes)
        node_to_qubit = defaultdict(int)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        circ = self.get_QAOA_circuit(self._p, self._origin_graph, node_to_qubit)  #, params
        state = QuantumState(len(self._origin_graph))
        state.set_zero_state()
        circ.update_quantum_state(state)
        state_samplings = state.sampling(1024)

        counts = defaultdict(int)
        for i in state_samplings:
            counts[i] += 1

        return counts

    def optimization_visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="qulacs")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()

    def sampling_visualization(self, counts):
        state_num = pow(2, len(self._origin_graph))
        state_counts = [counts[i] for i in range(state_num)]
        plt.figure()
        plt.bar(range(state_num), state_counts)
        plt.ylabel("Count number")
        plt.xlabel("State")
        plt.legend()
        plt.show()



# test
# from qulacs import Observable, QuantumCircuit, QuantumState
# from qulacs.gate import Y,CNOT,merge
#
# state = QuantumState(3)
# state.set_Haar_random_state()
#
# circuit = QuantumCircuit(3)
# circuit.add_X_gate(0)
# merged_gate = merge(CNOT(0,1),Y(1))
# circuit.add_gate(merged_gate)
# circuit.add_RX_gate(1,0.5)
# circuit.update_quantum_state(state)
#
# observable = Observable(3)
# observable.add_operator(2.0, "X 2 Y 1 Z 0")
# observable.add_operator(-3.0, "Z 2")
# value = observable.get_expectation_value(state)
# print(value)


import json
import networkx as nx
from queue import PriorityQueue
import matplotlib.pyplot as plt
import copy
import math
from collections import defaultdict
import os
import re
from quafu import Task
import numpy as np


class BuildLibrary:
    def __init__(self, backend='ScQ-P10', fidelity_threshold=95):
        self.backend = backend
        self.fidelity_threshold = fidelity_threshold
        task = Task()
        # task.load_account()
        task.config(backend=self.backend)
        self.backend_info = task.get_backend_info()
        self.calibration_time = self.backend_info['full_info']["calibration_time"]

    def get_structure(self):
        """
        Args:
            backend (str): topology data file. 'structure_ScQ-P20.txt'
        Returns:
            int_to_qubit(dict): {0: Q01, 1: Q02, ...}
            qubit_to_int(dict): {Q01: 0, Q02: 0, ...}
            directed_weighted_edges(list):[[qubit1,qubit2,fidelity],...]
            connected_substructure_list(list): [networkx.Graph,...]
        """

        json_topo_struct = self.backend_info['full_info']["topological_structure"]

        # # not ordered
        # qubits_list = []
        # for gate in json_topo_struct.keys():
        #     qubit = gate.split('_')
        #     if qubit[0] not in qubits_list:
        #         qubits_list.append(qubit[0])
        #     if qubit[1] not in qubits_list:
        #         qubits_list.append(qubit[1])

        # # ordered
        qubits_list = []
        for gate in json_topo_struct.keys():
            qubit = gate.split('_')
            qubits_list.append(qubit[0])
            qubits_list.append(qubit[1])
        qubits_list = list(set(qubits_list))
        qubits_list = sorted(qubits_list, key=lambda x: int(re.findall(r"\d+", x)[0]))

        int_to_qubit = {k: v for k, v in enumerate(qubits_list)}
        qubit_to_int = {v: k for k, v in enumerate(qubits_list)}
        directed_weighted_edges = []
        weighted_edges = []
        edges_dict = {}
        for gate, name_fidelity in json_topo_struct.items():
            gate_qubit = gate.split('_')
            qubit1 = qubit_to_int[gate_qubit[0]]
            qubit2 = qubit_to_int[gate_qubit[1]]
            gate_name = list(name_fidelity.keys())[0]
            fidelity = name_fidelity[gate_name]['fidelity']
            directed_weighted_edges.append([qubit1, qubit2, fidelity])
            gate_reverse = gate.split('_')[1] + '_' + gate.split('_')[0]
            if gate not in edges_dict and gate_reverse not in edges_dict:
                edges_dict[gate] = fidelity
            else:
                if fidelity < edges_dict[gate_reverse]:
                    edges_dict.pop(gate_reverse)
                    edges_dict[gate] = fidelity

        for gate, fidelity in edges_dict.items():
            gate_qubit = gate.split('_')
            qubit1, qubit2 = qubit_to_int[gate_qubit[0]], qubit_to_int[gate_qubit[1]]
            weighted_edges.append([qubit1, qubit2, fidelity])

        G = nx.Graph()
        G.add_weighted_edges_from(weighted_edges)
        G0 = copy.deepcopy(G)
        connected_substructure_list = []
        while len(G.nodes()) > 0:
            connected_nodes = max(nx.connected_components(G), key=len)
            connected_subgraph = G0.subgraph(connected_nodes)
            connected_substructure_list.append(connected_subgraph)
            G.remove_nodes_from(connected_nodes)
            # # draw coupling graph
            plt.close()
            # pos = nx.spring_layout(connected_subgraph)
            # new_labels = dict(map(lambda x: ((x[0], x[1]), format(float(str(x[2]['weight'])), '.2f')),
            #                       connected_subgraph.edges(data=True)))
            # nx.draw_networkx_edge_labels(connected_subgraph, pos=pos, edge_labels=new_labels, font_size=8)
            # nx.draw(connected_subgraph, pos=pos, with_labels=True, node_color='r', node_size=150, edge_color='b',
            #         width=1, font_size=9)
            # plt.show()
        return int_to_qubit, qubit_to_int, directed_weighted_edges, connected_substructure_list

    def substructure(self, directed_weighted_edges, connected_substructure_list, qubits_need):
        all_substructure = []
        substructure_nodes = []
        while len(all_substructure) == 0:
            for cg in connected_substructure_list:
                if len(cg.nodes()) >= qubits_need:
                    sorted_edges = sorted(cg.edges(data=True), key=lambda x: x[2]['weight'], reverse=True)
                    for elem in sorted_edges:
                        if elem[2]['weight'] > self.fidelity_threshold:
                            neighbors = PriorityQueue()
                            neighbors.put((-1, elem[0]))
                            ret_nodes = []
                            log_weight_product = 0
                            for node in cg.nodes():
                                cg.nodes[node]['visited'] = False
                            while neighbors.not_empty:
                                temp = neighbors.get()
                                node = temp[1]
                                if cg.nodes[node]['visited']:
                                    continue
                                weight = -temp[0]
                                log_weight_product += math.log(weight)
                                cg.nodes[node]['visited'] = True
                                ret_nodes.append(node)
                                if len(ret_nodes) == qubits_need:
                                    break
                                for neighbor in cg[node]:
                                    if not cg.nodes[neighbor]['visited']:
                                        weight = cg[node][neighbor]['weight']
                                        neighbors.put((-weight, neighbor))
                            out = []
                            for edge in directed_weighted_edges:
                                if edge[0] in ret_nodes and edge[1] in ret_nodes:
                                    out.append(edge)
                            if sorted(ret_nodes) not in substructure_nodes and all(
                                    qubit[2] > self.fidelity_threshold for qubit in out):
                                substructure_nodes.append(sorted(ret_nodes))
                                all_substructure.append([log_weight_product, out])

            self.fidelity_threshold = self.fidelity_threshold - 1
        all_substructure = sorted(all_substructure, key=lambda x: x[0], reverse=True)
        return all_substructure

    def build_substructure_library(self):
        # substructure_dict = defaultdict(list)
        substructure_dict = {}
        int_to_qubit, qubit_to_int, directed_weighted_edges, connected_substructure_list = self.get_structure()
        substructure_dict[1] = [[[q, q, 0.99]] for q in int_to_qubit.keys()]
        for qubits in range(2, len(connected_substructure_list[0].nodes()) + 1):
            sub_graph = self.substructure(directed_weighted_edges, connected_substructure_list, qubits)
            qlist = []
            for j in range(len(sub_graph)):
                # substructure_dict[qubits].append(sub_graph[j][1])
                qlist.append(sub_graph[j][1])
            substructure_dict[qubits] = qlist

        sorted_weighted_edges = sorted(directed_weighted_edges, key=lambda x: x[2], reverse=True)
        save_substructure = {'calibration_time': self.calibration_time, 'structure': sorted_weighted_edges, 'substructure_dict': substructure_dict,
                             'int_to_qubit': int_to_qubit, 'qubit_to_int': qubit_to_int}
        # if os.path.exists('LibSubstructure_' + self.backend + '.txt'):
        #     os.remove('LibSubstructure_' + self.backend + '.txt')
        with open('LibSubstructure_' + self.backend + '.txt', 'w') as file:
            # file.write(json.dumps(save_substructure))
            file.write(str(save_substructure))
        return save_substructure

    def chain_library_2D(self, substructure_data):
        # find one-dimensional chain
        G = nx.DiGraph()
        G.add_weighted_edges_from(substructure_data['substructure_dict'][len(substructure_data['qubit_to_int'])][0])
        node_degree = dict(G.degree)
        sort_degree = sorted(node_degree.items(), key=lambda kv: kv[1], reverse=False)
        one_link_nodes = [node for node, degree in sort_degree if degree == 2]
        all_oneD_chain = []
        for i in range(len(one_link_nodes)):
            begin_node = one_link_nodes[i]
            queue_oneD_chain = []
            queue_oneD_chain.append([begin_node])

            while queue_oneD_chain:
                end_node = queue_oneD_chain[0][-1]
                neighbor_nodes = [k for k, v in G[end_node].items()]
                neighbor_nodes = [node for node in neighbor_nodes if node not in queue_oneD_chain[0]]
                if len(neighbor_nodes) == 0:
                    all_oneD_chain.append(queue_oneD_chain[0])
                    queue_oneD_chain.remove(queue_oneD_chain[0])
                elif len(neighbor_nodes) == 1:
                    queue_oneD_chain[0].append(neighbor_nodes[0])
                else:
                    for k in range(1, len(neighbor_nodes)):
                        oneD_chain = copy.deepcopy(queue_oneD_chain[0])
                        oneD_chain.append(neighbor_nodes[k])
                        if oneD_chain not in queue_oneD_chain:
                            queue_oneD_chain.append(oneD_chain)
                    queue_oneD_chain[0].append(neighbor_nodes[0])

        chain_dict = defaultdict(list)

        sorted_chain = []
        for chain in all_oneD_chain:
            if sorted(chain) not in sorted_chain:
                sorted_chain.append(sorted(chain))
                chain_dict[len(chain)].append(chain)

        chain_dict = dict(sorted(chain_dict.items(), key=lambda x: x[0]))
        # longset_chain = chain_dict[max(chain_dict.keys())][0]

        structure_dict = {}
        for edge in substructure_data['structure']:
            structure_dict[(edge[0], edge[1])] = edge[2]

        connected_chain_list = []
        for node_number, chain_nodes_list in chain_dict.items():
            for chain_nodes in chain_nodes_list:
                chain_graph = nx.DiGraph()
                directed_weighted_edges = []
                for i in range(len(chain_nodes)-1):
                    directed_weighted_edges.append([chain_nodes[i], chain_nodes[i+1],
                                                    structure_dict[(chain_nodes[i], chain_nodes[i+1])]])
                    directed_weighted_edges.append([chain_nodes[i+1], chain_nodes[i],
                                                    structure_dict[(chain_nodes[i+1], chain_nodes[i])]])
                chain_graph.add_weighted_edges_from(directed_weighted_edges)
                connected_chain_list.append([directed_weighted_edges, chain_graph])

        subchain_dict = defaultdict(list)
        for chain in connected_chain_list:
            for qubits in range(2, len(chain[1].nodes())+1):
                sub_graph = self.substructure(chain[0], [chain[1]], qubits)
                for j in range(len(sub_graph)):
                    log_weight_product = 0
                    for edge in sub_graph[j][1]:
                        log_weight_product = log_weight_product + math.log(edge[2])
                    if [log_weight_product, sub_graph[j][1]] not in subchain_dict[qubits]:
                        subchain_dict[qubits].append([log_weight_product, sub_graph[j][1]])

        sorted_subchain_dict = {}
        for k, chain_list in subchain_dict.items():
            chain_list = sorted(chain_list, key=lambda x: x[0], reverse=True)
            chain_list = [chain[1] for chain in chain_list]
            sorted_subchain_dict[k] = chain_list

        save_substructure = {'calibration_time': self.calibration_time, 'subchain_dict': sorted_subchain_dict}
        if os.path.exists('LibSubchain_' + self.backend + '.txt'):
            os.remove('LibSubchain_' + self.backend + '.txt')
        with open('LibSubchain_' + self.backend + '.txt', 'w') as file:
            file.write(str(save_substructure))
        return save_substructure




# pylint: disable=invalid-name,wrong-import-position

"""Main QCover public functionality."""

import pkgutil

# Allow extending this namespace. Please note that currently this line needs
# to be placed *before* the wrapper imports or any non-import code AND *before*
# importing the package you want to allow extensions for (in this case `backends`).
__path__ = pkgutil.extend_path(__path__, __name__)
__version__ = '2.3.0'
__license__ = 'Apache-2.0 License'

from Qcover.applications import *
from Qcover.backends import *
from Qcover.optimizers import *





import qiskit

from qiskit import *

from qiskit import Aer

from Qcover.core import Qcover

from typing import Optional

from Qcover.backends import Backend,CircuitByQulacs

from Qcover.optimizers import Optimizer,COBYLA

from Qcover.utils import get_graph_weights, generate_weighted_graph, generate_graph_data

import networkx as nx

import numpy as np

import matplotlib.pyplot as plt

import warnings
warnings.filterwarnings("ignore")

class QAOA_OTOC:
    
    def __init__(self,
                 p: int = 1,
                 graph: nx.Graph = None,
                 optimizer: Optional[Optimizer] = COBYLA(),
                 backend: Optional[Backend] = CircuitByQulacs(),
                 ) -> None:
        
        assert graph is not None
        self._p = p
        self._original_graph = graph

        self._qc = Qcover(self._original_graph,
                          self._p,
                          optimizer=optimizer,
                          backend=backend,
                          research_obj="QAOA")

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, ap):
        self._p = ap

    @property
    def qc(self):
        return self._qc

    @property
    def original_graph(self):
        return self._original_graph

    @original_graph.setter
    def original_graph(self, graph):
        """
        according to the type of graph(nx.graph / tuple) to set the value of
        self._original_graph
        """
        if isinstance(graph, nx.Graph):
            self._original_graph = graph
        elif isinstance(graph, tuple):
            assert (len(graph) >= 2) and (len(graph) <= 3)

            if len(graph) == 2:
                node_num, edge_num = graph
                wr = None
            elif len(graph) == 3:
                node_num, edge_num, wr = graph

            nodes, edges = generate_graph_data(node_num, edge_num, wr)
            self._original_graph = generate_weighted_graph(nodes, edges)
        elif isinstance(graph, list):
            assert len(graph) == 3
            node_list, edge_list, weight_range = graph
            self._original_graph = generate_weighted_graph(node_list, edge_list, weight_range)
        else:
            print("Error: the argument graph should be a instance of nx.Graph "
                  "or a tuple formed as (node_num, edge_num)")


    def run(self, is_parallel=False):
        res = self._qc.run(is_parallel=is_parallel)  # True
        return res
    
    def scrambling_circuit(self, G, butterfly, theta):
        p=int(len(theta)/2)
        N_qubits = len(G.nodes())+1
        ctrbit=1
        qc = QuantumCircuit(N_qubits,ctrbit)
    
        gamma = theta[:p]
        beta = theta[p:]
    
        #Ctrbit#
        ctr=0
        qc.h(ctr)
        qc.p(-np.pi/2, ctr)
    
        #N qubits#
        for j in range(1, N_qubits):
            qc.h(j)
            
        qc.cz(ctr,1)
        qc.barrier()
    
        ### U transformation ###
        for k in range(p):
            for i in range(1, N_qubits):
                qc.rz(2 * gamma[k]*nx.get_node_attributes(G, "weight")[i],i)
        
            for pair in list(G.edges()):
                qc.rzz(2*gamma[k]*nx.get_edge_attributes(G, "weight")[(pair[0],pair[1])], pair[0], pair[1])
        
            for nd in range(1, N_qubits):
                qc.rx(2*beta[k], nd)
                
        ### perturbation ###
        qc.barrier()
        qc.x(butterfly)
        qc.barrier()

        #### U_dag transformation #####
        for k_dag in range(p):
            for nd_dag in range(1, N_qubits):
                qc.rx(-2*beta[p-1-k_dag], nd_dag)

            for pair_dag in list(G.edges()):
                qc.rzz(-2*gamma[p-1-k_dag]*nx.get_edge_attributes(G, "weight")[(pair_dag[0],pair_dag[1])], pair_dag[0], pair_dag[1])
        
            for i_dag in range(1, N_qubits):
                qc.rz(-2 * gamma[p-1-k_dag]*nx.get_node_attributes(G, "weight")[i_dag], i_dag)

        qc.barrier()
        qc.cz(ctr,1)
        
        ###measure the ctrl bit###
        qc.p(-np.pi/2, ctr)
        qc.h(ctr)
        qc.measure(0, 0)
    
        return qc
    
    def calculate_OTOC(self, G, butterfly, theta):
        
        scric=self.scrambling_circuit(G, butterfly, theta)
        backend = Aer.get_backend('qasm_simulator')
        shots = 10000
        results = execute(scric, backend=backend, shots=shots).result()
        answer = results.get_counts()
        if answer['1'] is None:
            pass
        elif answer['1']==shots:
            measure=1
        else:
            measure=(answer['1']-answer['0'])/shots
      
        return measure
    
    
if __name__ == '__main__':
    p = 10  #
    g = nx.Graph()
    ########！！！！because there is an auxiliary qubit No.0, we need to start the No. of graphs with 1.
    nodes = [(1, 0), (2, 0), (3, 0),(4, 0),(5, 0),(6, 0),(7, 0)]
    edges = [(1, 2, -1), (2, 3, -1),(3, 4, -1),(4, 5, -1),(5, 6, -1),(6, 7, -1)]

    for nd in nodes:
        u, w = nd[0], nd[1]
        g.add_node(int(u), weight=int(w))
    for ed in edges:
        u, v, w = ed[0], ed[1], ed[2]
        g.add_edge(int(u), int(v), weight=int(w))

    ##############  
    qcover_bc = CircuitByQulacs()
    optc = COBYLA(options={'disp': True,'tol': 1e-3, 'maxiter': 10000}) #

    qaoa_otoc = QAOA_OTOC(graph=g,
                        p=p,
                        optimizer=optc,
                        backend=qcover_bc) 

    res = qaoa_otoc.run()

    print("solution is:", res)
        
    params = res["Optimal parameter value"]

    ###########dynamics of OTOC in the circuit
    a=[]
    b=[]
    OTOC_data=[]
        
    for l in range(p):
        a.append(params[l])
        b.append(params[p+l])
        cta=a+b

        OTOC_measure = qaoa_otoc.calculate_OTOC(g,4,cta)
        OTOC_data.append(OTOC_measure)

    
    xlabel='depth'
    ylabel='OTOC value'
    plt.figure(figsize=(13,6))

    x=[]
    for i in range(p):
        x.append(i+1)

    y=OTOC_data

    plt.plot(x,y)
    plt.xlabel(xlabel, fontsize=14)
    plt.ylabel(ylabel, fontsize=14)
    plt.show()



# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import time
import logging
import random
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_regular_graph


logger = logging.getLogger(__name__)


class GraphColoring:
    """
    graph coloring problem:
    Given a graph and K colors, each vertex is required to have only one color,
    and the colors of the connected vertices are different
    """
    def __init__(self,
                 graph: nx.Graph = None,
                 node_num: int = None,
                 color_num: int = 4,
                 node_degree: int = 3,
                 weight_range: int = 1,
                 seed: int = None,
                 penalty: int = 99998
                 ):

        self._color_num = color_num
        self._P = penalty
        if graph is None:
            self._node_num = node_num
            # self._node_degree = node_degree
            # self._weight_range = weight_range
            # self._seed = seed
            self._graph = random_regular_graph(node_num=self._node_num,
                                               degree=node_degree,
                                               weight_range=weight_range,
                                               seed=seed)   # self.get_random_regular_graph()
        else:
            self._node_num = len(graph.nodes)
            self._graph = graph
            
        self._qmatrix = None
        self._shift = None

    @property
    def node_num(self):
        return self._node_num

    @property
    def color_num(self):
        return self._color_num

    @property
    def graph(self):
        return self._graph

    def update_random_graph(self, node_num, color_num, node_degree, weight_range, seed):
        self._node_num = node_num
        self._color_num = color_num

        self._graph = random_regular_graph(node_num=self._node_num,
                                           degree=node_degree,
                                           weight_range=weight_range,
                                           seed=seed)

    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of max cut problem
        Args:
            adjacent_mat (np.array):  the adjacent matrix of graph G of the problem
            cnt (int): number of vertices in the graph of the problem
            k: the number of colors that can be used
            P: an arbitrarily large number used as constant parameter

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        """

        Qmat_length = self._node_num * self._color_num
        qubo_mat = -self._P * np.eye(Qmat_length)
        adj_mat = nx.adjacency_matrix(self._graph).A
        for i in range(Qmat_length):
            for j in range(Qmat_length):
                x, y = i // self._color_num, j // self._color_num
                idl, idr = x * self._color_num, (x + 1) * self._color_num - 1
                if (i >= idl and i <= idr) and (j >= idl and j <= idr):
                    if i == j:
                        qubo_mat[i][i] = -self._P
                    else:
                        qubo_mat[i][j] = self._P
                else:
                    if adj_mat[x][y] != 0:
                        for l in range(self._color_num):
                            qubo_mat[x*self._color_num + l][y*self._color_num + l] = self._P/2
               
        shift = self._P
        return qubo_mat, shift

    def run(self):
        if self._qmatrix is None:
            self._qmatrix, self._shift = self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        gc_graph = get_weights_graph(ising_mat)
        return gc_graph, self._shift
    


import numpy as np
import networkx as nx
from typing import Optional
from collections import defaultdict

from Qcover.utils import get_graph_weights, generate_weighted_graph, generate_graph_data
from Qcover.optimizers import Optimizer, COBYLA
from Qcover.backends import Backend, CircuitByQiskit
from Qcover.core import Qcover
import warnings
warnings.filterwarnings("ignore")


class QAOA_Generate:
    def __init__(self,
                 graph: nx.Graph = None,
                 p: int = 1,
                 optimizer: Optional[Optimizer] = COBYLA(),
                 backend: Optional[Backend] = CircuitByQiskit(),
                 ) -> None:

        assert graph is not None
        self._p = p
        self._original_graph = graph
        self._qc = Qcover(self._original_graph,
                          self._p,
                          optimizer=optimizer,
                          backend=backend,
                          research_obj="QAOA")

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, ap):
        self._p = ap

    @property
    def qc(self):
        return self._qc

    @property
    def original_graph(self):
        return self._original_graph

    @original_graph.setter
    def original_graph(self, graph):
        """
        according to the type of graph(nx.graph / tuple) to set the value of
        self._original_graph
        """
        if isinstance(graph, nx.Graph):
            self._original_graph = graph
        elif isinstance(graph, tuple):
            assert (len(graph) >= 2) and (len(graph) <= 3)

            if len(graph) == 2:
                node_num, edge_num = graph
                wr = None
            elif len(graph) == 3:
                node_num, edge_num, wr = graph

            nodes, edges = generate_graph_data(node_num, edge_num, wr)
            self._original_graph = generate_weighted_graph(nodes, edges)
        elif isinstance(graph, list):
            assert len(graph) == 3
            node_list, edge_list, weight_range = graph
            self._original_graph = generate_weighted_graph(node_list, edge_list, weight_range)
        else:
            print("Error: the argument graph should be a instance of nx.Graph "
                  "or a tuple formed as (node_num, edge_num)")

    def run(self, is_parallel=False):
        # nx.draw(self._g)
        # if self._qc is None:
        #     self._qc = Qcover(self._g,
        #                       self._p,
        #                       optimizer=self._optimizer,
        #                       backend=self._backend,
        #                       research_obj="GHZ")
        res = self._qc.run(is_parallel=is_parallel)  # True
        return res


if __name__ == '__main__':
    p = 3
    g = nx.Graph()
    nodes = [(0, 0), (1, 0), (2, 0)]
    edges = [(0, 1, 1), (1, 2, 1)]

    for nd in nodes:
        u, w = nd[0], nd[1]
        g.add_node(int(u), weight=int(w))
    for ed in edges:
        u, v, w = ed[0], ed[1], ed[2]
        g.add_edge(int(u), int(v), weight=int(w))

    from Qcover.optimizers import COBYLA, Interp, SLSQP, L_BFGS_B, GradientDescent, SPSA, Fourier
    optf = Fourier(p=p, q=4, r=3)
    optc = COBYLA(options={'maxiter': 300, 'disp': True, 'rhobeg': 1.0, 'tol': 1e-12}) #, initial_point=opti_params
    opti = Interp(optimize_method="COBYLA", options={'tol': 1e-12, 'disp': False}) #, initial_point=opti_params, "optimal_value": -2*node_num, "approximate_ratio": 0.9
    opts = SLSQP(options={'maxiter': 100,
                          'ftol': 1e-06,
                          'iprint': 1,
                          'disp': False,
                          'eps': 1.4901161193847656e-08,
                          'finite_diff_rel_step': None})

    optl = L_BFGS_B(options={'disp': None,
                             'maxcor': 10,
                             'ftol': 2.220446049250313e-09,
                             'gtol': 1e-05,
                             'eps': 1e-08,
                             'maxfun': 15000,
                             'maxiter': 15000,
                             'iprint': -1,
                             'maxls': 20,
                             'finite_diff_rel_step': None})  #

    optg = GradientDescent(options={'maxiter':300, 'learning_rate':0.05, 'tol':1e-6})
    opta = SPSA(options={'A':34, 'R': 15, 'maxiter':300, 'tol':1e-6})
    from Qcover.backends import CircuitByQiskit, CircuitByQton, CircuitByQulacs, CircuitByProjectq, CircuitByTensor, CircuitByCirq

    qiskit_bc = CircuitByQiskit(expectation_calc_method="statevector")
    qulacs_bc = CircuitByQulacs()

    # cirq_bc = CircuitByCirq()
    # projectq_bc = CircuitByProjectq()   # Bugs need to fix
    # ts = CircuitByTensor()
    # qt = CircuitByQton()  #expectation_calc_method="tensor"

    qaoa = QAOA_Generate(graph=g,
                           p=p,
                           optimizer=optc,
                           backend=qiskit_bc)  #qt, , cirq_bc, projectq_bc, qulacs_bc
    res = qaoa.run()

    print("solution is:", res)
    params = res["Optimal parameter value"]
    # params = [2.40756694, 2.37257408, 2.34787417, 0.79040416, 0.78994118,
    #    0.73654257]
    qaoa._qc.backend._pargs = params

    res_exp = qaoa._qc.backend.expectation_calculation()
    print("optimal parameter value:", qaoa._qc.backend._pargs)
    print("the optimal expectation is: ", res_exp)
    out_count = qaoa._qc.backend.get_result_counts(params)
    import matplotlib.pyplot as plt
    from qiskit.visualization import plot_histogram
    plot_histogram(out_count)
    plt.show()




import sys
from typing import Optional
import logging
import numpy as np
from scipy import optimize as opt
from scipy.optimize import shgo
from Qcover.optimizers import Optimizer
from Qcover.exceptions import ArrayShapeError

logger = logging.getLogger(__name__)


class SHGO(Optimizer):
    """
    SLSQP: a numerical optimization method for constrained problems

    based on scipy.optimize.minimize SLSQP.
    For further detail, please refer to
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
    """

    # pylint: disable=unused-argument
    def __init__(self,
                 options: dict = None,
                 initial_point: Optional[np.ndarray] = None) -> None:
        """
        Args:
            options: some optional setting parameters such as:
                maxiter: Maximum number of function evaluations.
                disp: Set to True to print convergence messages.
                rhobeg: Reasonable initial changes to the variables.
                tol: Final accuracy in the optimization (not precisely guaranteed).
                     This is a lower bound on the size of the trust region.
        """
        super().__init__()
        self._p = None
        self._options = options
        self._initial_point = initial_point

    def optimize(self, objective_function):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * self._p)])
        else:
            try:
                if len(self._initial_point) != 2 * self._p:
                    raise ArrayShapeError("The shape of initial parameters is not match with p")
            except ArrayShapeError as e:
                print(e)
                sys.exit()

        # method = self._options['method'] if 'method' in self._options['method'] else 'SLSQP'
        sampling_method = self._options['sampling_method'] if 'sampling_method' in self._options else 'simplicial'
        minimizer_kwargs = self._options['minimizer_kwargs'] if 'minimizer_kwargs' in self._options else None
        bounds = [(None, None), ] * len(self._initial_point)
        res = shgo(objective_function, bounds=bounds, sampling_method=sampling_method, minimizer_kwargs=minimizer_kwargs)
        # res = opt.minimize(objective_function,
        #                    x0=np.array(self._initial_point),
        #                    args=self._p,
        #                    method='SLSQP',
        #                    jac=None,
        #                    options=self._options
        #                    )

        return res.x, res.fun, res.nfev

# opts = SLSQP(options={'maxiter': 100,
#         'ftol': 1e-06,
#         'iprint': 1,
#         'disp': False,
#         'eps': 1.4901161193847656e-08,
#         'finite_diff_rel_step': None})


import itertools
import os
from collections import defaultdict
import matplotlib.pyplot as plt
from multiprocessing import Pool, cpu_count
import networkx as nx
import cirq
from Qcover.backends import Backend
from Qcover.utils import get_graph_weights


class CircuitByCirq(Backend):
    """generate a instance of CircuitByCirq"""

    def __init__(self,
                 research: str = "QAOA",
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByCirq"""
        super(CircuitByCirq, self).__init__()
        self._p = None
        self._origin_graph = None
        self._is_parallel = False if is_parallel is None else is_parallel
        self._research = research

        self._nodes_weight = None
        self._edges_weight = None
        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []
        self._element_expectation = dict()

    @property
    def element_expectation(self):
        return self._element_expectation

    def get_operator(self, element, qubit_num):
        qubits = cirq.LineQubit.range(qubit_num)
        nodes = [element] if isinstance(element, int) else list(element)
        op = 1
        for i in range(qubit_num):
            if i in nodes:
                op *= cirq.Z(qubits[i])
        return op

    def get_QAOA_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ = cirq.Circuit()
        ql = cirq.LineQubit.range(len(graph))
        circ.append(cirq.H(ql[i]) for i in range(len(graph)))
        for k in range(p):
            for i in graph.nodes:
                u = node_to_qubit[i]
                circ.append(cirq.rz(2 * gamma_list[k] * self._nodes_weight[i]).on(ql[u]))

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue

                circ.append(cirq.CX(ql[u], ql[v]))
                circ.append(cirq.rz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]]).on(ql[v]))
                circ.append(cirq.CX(ql[u], ql[v]))

            for nd in graph.nodes:
                u = node_to_qubit[nd]
                circ.append(cirq.Moment(cirq.rx(2 * beta_list[k]).on(ql[u])))

        return circ

    def get_GHZ_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ = cirq.Circuit()
        ql = cirq.LineQubit.range(len(graph))
        circ.append(cirq.H(ql[i]) for i in range(len(graph)))

        pivot = len(self._origin_graph) // 2
        for k in range(p):
            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue

                circ.append(cirq.CX(ql[u], ql[v]))
                circ.append(cirq.rz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]]).on(ql[v]))
                circ.append(cirq.CX(ql[u], ql[v]))

            for nd in graph.nodes:
                if nd != pivot:
                    u = node_to_qubit[nd]
                    circ.append(cirq.Moment(cirq.rx(2 * beta_list[k]).on(ql[u])))

        return circ

    def get_expectation(self, element_graph, p=None):
        """
        transform the graph to circuit according to the computing_framework
        Args:
            graph (nx.Graph): graph to be transformed to circuit
            params (np.array): Optimal parameters
            original_e (Optional[None, int, tuple])
        Return:
            if original_e=None, then the graph is the whole original graph generated by
            generate_weighted_graph(), so just return the circuit transformed by it

            if original_e is a int, then the subgraph is generated by node(idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[], and the circuit

            if original_e is a tuple, then the subgraph is generated by edge(node idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[] as
            tuple(mapped node_id1, mapped node_id2), and the circuit
        """
        if self._is_parallel is False:
            p = self._p if p is None else p
            original_e, graph = element_graph
        else:
            p = self._p if len(element_graph) == 1 else element_graph[1]
            original_e, graph = element_graph[0]

        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        if self._research == "QAOA":
            circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
        elif self._research == "GHZ":
            pivot = len(self._origin_graph) // 2
            if pivot in graph:
                circ = self.get_GHZ_circuit(p, graph, node_to_qubit)
            else:
                circ = self.get_QAOA_circuit(p, graph, node_to_qubit)

        qubits = cirq.LineQubit.range(len(node_list))
        qubit_map = dict(zip(qubits, range(len(node_list))))

        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            op = cirq.Z(qubits[node_to_qubit[original_e]])
        else:
            weight = self._edges_weight[original_e]
            op = cirq.Z(qubits[node_to_qubit[original_e[0]]]) * cirq.Z(qubits[node_to_qubit[original_e[1]]])

        state = cirq.final_state_vector(circ)
        exp_res = op.expectation_from_state_vector(state, qubit_map=qubit_map)

        return weight, exp_res.real

    def expectation_calculation(self, p=None):
        if self._nodes_weight is None or self._edges_weight is None:
            nodes_weight, edges_weight = get_graph_weights(self._origin_graph)
            self._nodes_weight, self._edges_weight = nodes_weight, edges_weight

        self._element_expectation = {}
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        cpu_num = cpu_count()
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        res = 0
        for item in self._element_to_graph.items():
            w_i, exp_i = self.get_expectation(item, p)
            if isinstance(item[0], tuple):
                self._element_expectation[item[0]] = exp_i
            res += w_i * exp_i

        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        args = list(itertools.product(self._element_to_graph.items(), [p]))
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, args))

        pool.terminate()  # pool.close()
        pool.join()
        res = 0
        for it in circ_res[0]:
            res += it[0] * it[1]

        # res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def optimization_visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="cirq")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()



import networkx as nx
# import itertools
# import numpy as np
import random
from qiskit import QuantumCircuit, QuantumRegister
import math
from collections import defaultdict
import copy
import re
import matplotlib.pyplot as plt
import os
from .hardware_library import BuildLibrary
from quafu import User, Task
from quafu import QuantumCircuit as quafuQC


class CompilerForQAOA:
    """
    compilerforQAOA is used to compile the combinatorial optimization problem weight graph
    into a quantum hardware executable circuit and send it to the quafu cloud platform for execution.
    Args:
        g (networkx.Graph): Weight graph created by networkx.
        p (int): QAOA algorithm layers
        optimal_params (list): Optimal parameters of QAOA circuit calculated by Qcover.
        apitoken (str): API token of your quafu account,
                        you can get it on the quafu quantum computing cloud platform (http://quafu.baqis.ac.cn/).
        cloud_backend (str): Currently you can choose from three quantum chips: 'ScQ-P10', 'ScQ-P20', 'ScQ-P50'
    """

    def __init__(self,
                 g: nx.Graph = None,
                 p: int = 1,
                 optimal_params: list = None,
                 apitoken: str = None,
                 cloud_backend: str = "ScQ-P10") -> None:
        self._gate = "CNOT"
        self._g = g
        self._p = p
        self._optimal_params = optimal_params
        self._apitoken = apitoken
        self._cloud_backend = cloud_backend
        self._nodes = {(k, v['weight']) for k, v in g.nodes()._nodes.items()}
        self._edges = {(u, v, g.adj[u][v]['weight']) for u, v in list(g.edges())}
        self._logical_qubits = len(g.nodes())
        self._physical_qubits = self._logical_qubits
        self._logi2phys_mapping = {}

    def random_layout_mapping(self):
        """
        Args:
            physical_qubits (int): The number of hardware qubits
            logical_qubits (int): The number of qubits in a circuit
        Returns:
            qubits_mapping (dict): Random mapping of logical qubits to physical qubits
        """
        if self._physical_qubits >= self._logical_qubits:
            random_physical_qubits = random.sample(range(0, self._physical_qubits), self._logical_qubits)
            qubits_mapping = {}
            for i in range(0, self._logical_qubits):
                qubits_mapping[random_physical_qubits[i]] = i
            return qubits_mapping
        else:
            print("Error: physical qubits must be larger than logical qubits")

    def simple_layout_mapping(self, phys_qubits_order=None):
        """
        Args:
            phys_qubits_order (list): The order of the selected physical qubits
        Returns:
            qubits_mapping (dict): Simple mapping of logical qubits to physical qubits
        """
        if self._physical_qubits >= self._logical_qubits:
            phys_qubits_order = [i for i in
                                 range(0, self._logical_qubits)] if phys_qubits_order is None else phys_qubits_order
            qubits_mapping = {}
            for i in range(0, self._logical_qubits):
                qubits_mapping[i] = phys_qubits_order[i]
            return qubits_mapping
        else:
            print("Error: physical qubits must be larger than logical qubits")

    def QAOA_logical_circuit(self):
        lc = QuantumRegister(len(self._nodes), 'l')
        logical_circ = QuantumCircuit(lc)
        graph_edges = dict(
            [(tuple(sorted(list(self._edges)[i][0:2])), list(self._edges)[i][2]) for i in range(0, len(self._edges))])
        graph_nodes = dict(self._nodes)
        gamma, beta = self._optimal_params[:self._p], self._optimal_params[self._p:]
        for i in range(len(graph_nodes)):
            logical_circ.h(i)
        for k in range(0, self._p):
            # logical_circ.barrier()
            for u, v in graph_nodes.items():
                logical_circ.rz(2 * gamma[k] * v, u)
            for u, v in graph_edges.items():
                logical_circ.rzz(2 * gamma[k] * v, u[0], u[1])
            for u, v in graph_nodes.items():
                logical_circ.rx(2 * beta[k], u)
        return logical_circ

    def sorted_nodes_degree(self):
        """
        Returns:
            sort_nodes (np.array): nodes are sorted by node degree in descending order
        """
        node_degree = dict(self._g.degree)
        sort_degree = sorted(node_degree.items(), key=lambda kv: kv[1], reverse=True)
        sort_nodes = [sort_degree[i][0] for i in range(len(sort_degree))]
        return sort_nodes

    def scheduled_pattern_rzz_swap(self, qubits_mapping):
        """
        Get the fixed execution pattern of the QAOA circuit.
        Args:
            qubits_mapping (dict): {physical qubit: logical qubit}
                        example: {0: 1, 1: 2, 2: 0}
        Returns:
            pattern_rzz_swap (dict): {k: [[(q1,q2),(p1,p2)],[] ] ...},
                                    k-th execution cycle,
                                    execute rzz/swap gate between logic qubit q1 and q2 (physics qubit p1 and p2).
                                    gates execution pattern: rzz,rzz,swap,swap,rzz,rzz,swap,swap, ...
            rzz_gates_cycle (dict): rzz gate execution in k-th cycle, {(q1,q2): k, ...}
        """
        loop = 1
        cycle = 0
        pattern_rzz_swap = defaultdict(list)
        rzz_gates_cycle = defaultdict(list)
        qubits_mapping = dict(sorted(qubits_mapping.items(), key=lambda x: x[0]))
        mapping = qubits_mapping.copy()
        m = sorted(list(mapping.keys()))
        while loop <= math.ceil(self._logical_qubits / 2):
            r1 = 0
            while r1 < self._logical_qubits - 1:
                pattern_rzz_swap[cycle].append([(mapping[m[r1]], mapping[m[r1 + 1]]), (m[r1], m[r1 + 1])])
                rzz_gates_cycle[(mapping[m[r1]], mapping[m[r1 + 1]])] = [cycle, (m[r1], m[r1 + 1])]
                r1 = r1 + 2
            cycle = cycle + 1
            r2 = 1
            while r2 < self._logical_qubits - 1:
                pattern_rzz_swap[cycle].append([(mapping[m[r2]], mapping[m[r2 + 1]]), (m[r2], m[r2 + 1])])
                rzz_gates_cycle[(mapping[m[r2]], mapping[m[r2 + 1]])] = [cycle, (m[r2], m[r2 + 1])]
                r2 = r2 + 2
            cycle = cycle + 1
            if loop == math.ceil(self._logical_qubits / 2):
                break
            else:
                s1 = 1
                while s1 < self._logical_qubits - 1:
                    pattern_rzz_swap[cycle].append([(mapping[m[s1]], mapping[m[s1 + 1]]), (m[s1], m[s1 + 1])])
                    x = mapping[m[s1]]
                    mapping[m[s1]] = mapping[m[s1 + 1]]
                    mapping[m[s1 + 1]] = x
                    s1 = s1 + 2
                cycle = cycle + 1
                s2 = 0
                while s2 < self._logical_qubits - 1:
                    pattern_rzz_swap[cycle].append([(mapping[m[s2]], mapping[m[s2 + 1]]), (m[s2], m[s2 + 1])])
                    x = mapping[m[s2]]
                    mapping[m[s2]] = mapping[m[s2 + 1]]
                    mapping[m[s2 + 1]] = x
                    s2 = s2 + 2
                cycle = cycle + 1
            loop = loop + 1
        if self._logical_qubits % 2 == 1:
            pattern_rzz_swap.pop(cycle - 1)
            for item in pattern_rzz_swap[0]:
                rzz_gates_cycle[item[0]] = [0, item[1]]
        return pattern_rzz_swap, rzz_gates_cycle

    def best_initial_mapping(self, rzz_gates_cycle, truncation=5):
        simple_mapping = self.simple_layout_mapping()
        sorted_nodes = self.sorted_nodes_degree()
        rzz_gates_cycle = {tuple(sorted(k)): v[0] for k, v in rzz_gates_cycle.items()}
        mapping_logi2phys_list = []
        # mapping_logi2phys_list.append(({sorted_nodes[0]: 0}, 0))
        for n in range(len(sorted_nodes)):
            mapping_logi2phys_list.append(({sorted_nodes[0]: n}, 0))
        last_cycle = max(rzz_gates_cycle.values())
        # truncation = truncation
        # The larger the truncation, the more likely it is to find the optimal initial mapping,
        # but the time cost increases.
        for item in range(1, len(sorted_nodes)):
            node = sorted_nodes[item]
            update_mapping_list = []
            for mapping_path in mapping_logi2phys_list:
                q2p_path = mapping_path[0]
                used_physical_bits = list(q2p_path.values())
                used_logical_bits = list(q2p_path.keys())
                for phys_bit in range(self._logical_qubits):
                    deepest_cycle = mapping_path[1]
                    max_depth = 0
                    if phys_bit not in used_physical_bits:
                        for neighbor_node in self._g.neighbors(node):
                            if neighbor_node in used_logical_bits:
                                depth = rzz_gates_cycle[tuple(sorted([phys_bit, q2p_path[neighbor_node]]))]
                                if depth > max_depth:
                                    max_depth = depth
                        if max_depth < last_cycle:
                            q2p_path = copy.deepcopy(q2p_path)
                            q2p_path[node] = phys_bit
                            if max_depth > deepest_cycle:
                                deepest_cycle = max_depth
                            if deepest_cycle == 0:
                                update_mapping_list.append((q2p_path, deepest_cycle))
                            else:
                                if [depth for _, depth in update_mapping_list].count(deepest_cycle) < truncation:
                                    update_mapping_list.append((q2p_path, deepest_cycle))
            mapping_logi2phys_list = copy.deepcopy(update_mapping_list)

        mapping_logi2phys_list = sorted(mapping_logi2phys_list, key=lambda x: x[1])
        # print(mapping_logi2phys_list[0:50])
        # print(len(mapping_logi2phys_list))
        if mapping_logi2phys_list:
            best_phys2logi_mapping = {v: k for k, v in mapping_logi2phys_list[0][0].items()}
        else:
            best_phys2logi_mapping = simple_mapping
        return best_phys2logi_mapping

    def QAOA_physical_circuit(self, pattern_rzz_swap, qubits_mapping):
        """
        Get the fixed execution pattern of the QAOA circuit.
        Args:
            pattern_rzz_swap (dict): Execution pattern of rzz and swap gates
            qubits_mapping (dict): {physical bit (int): logical bit (int), ...}
        Returns:
            circuit: QAOA physical circuit #qiskit
            final_gates_scheduled (dict):
        """
        graph_edges = dict(
            [(tuple(sorted(list(self._edges)[i][0:2])), list(self._edges)[i][2]) for i in range(0, len(self._edges))])
        graph_nodes = dict(self._nodes)
        gamma, beta = self._optimal_params[:self._p], self._optimal_params[self._p:]
        mapping = qubits_mapping.copy()
        qubits_mapping_initial = qubits_mapping.copy()
        qubits_mapping_initial = dict(sorted(qubits_mapping_initial.items(), key=lambda x: x[0]))
        gates_scheduled = defaultdict(list)
        m = sorted(list(mapping.keys()))
        depth = 0
        for i in range(len(graph_nodes)):
            u = mapping[m[i]]
            gates_scheduled[depth].append(('Rz', (u, graph_nodes[u])))
        depth = depth + 1

        loop = 1
        cycle = 0
        while loop <= math.ceil(len(graph_nodes) / 2):
            r1 = 0
            for i in range(len(pattern_rzz_swap[cycle])):
                if tuple(sorted(pattern_rzz_swap[cycle][i][0])) in list(graph_edges.keys()):
                    edges_weight = graph_edges[tuple(sorted(pattern_rzz_swap[cycle][i][0]))]
                    gates_scheduled[depth + cycle].append(('Rzz', (pattern_rzz_swap[cycle][i][0], edges_weight)))
                    graph_edges.pop(tuple(sorted(pattern_rzz_swap[cycle][i][0])))
                r1 = r1 + 2
            cycle = cycle + 1
            if cycle == len(pattern_rzz_swap) or not graph_edges:
                break

            r2 = 1
            for i in range(len(pattern_rzz_swap[cycle])):
                if tuple(sorted(pattern_rzz_swap[cycle][i][0])) in list(graph_edges.keys()):
                    edges_weight = graph_edges[tuple(sorted(pattern_rzz_swap[cycle][i][0]))]
                    gates_scheduled[depth + cycle].append(('Rzz', (pattern_rzz_swap[cycle][i][0], edges_weight)))
                    graph_edges.pop(tuple(sorted(pattern_rzz_swap[cycle][i][0])))
                r2 = r2 + 2
            cycle = cycle + 1
            if cycle == len(pattern_rzz_swap) or not graph_edges:
                break

            s1 = 1
            for i in range(len(pattern_rzz_swap[cycle])):
                gates_scheduled[depth + cycle].append(('SWAP', pattern_rzz_swap[cycle][i][0]))
                x = mapping[m[s1]]
                mapping[m[s1]] = mapping[m[s1 + 1]]
                mapping[m[s1 + 1]] = x
                s1 = s1 + 2
            cycle = cycle + 1

            s2 = 0
            for i in range(len(pattern_rzz_swap[cycle])):
                gates_scheduled[depth + cycle].append(('SWAP', pattern_rzz_swap[cycle][i][0]))
                x = mapping[m[s2]]
                mapping[m[s2]] = mapping[m[s2 + 1]]
                mapping[m[s2 + 1]] = x
                s2 = s2 + 2
            cycle = cycle + 1
            loop = loop + 1

        depth = depth + cycle
        for i in range(len(graph_nodes)):
            gates_scheduled[depth].append(('Rx', (mapping[m[i]], 1)))

        first_gates_scheduled = defaultdict(list)
        layer_keys = sorted(list(gates_scheduled.keys()))
        for i in range(len(layer_keys)):
            first_gates_scheduled[i] = gates_scheduled[layer_keys[i]]

        # first_gates_scheduled: The first layer of circuit
        # final_gates_scheduled: Construct the entire QAOA circuit through the first layer of circuit
        final_gates_scheduled = defaultdict(list)
        a = len(first_gates_scheduled)
        depth = 0
        for i in range(len(graph_nodes)):
            mh = list(qubits_mapping_initial.keys())
            final_gates_scheduled[depth].append(('H', mh[i], qubits_mapping_initial[mh[i]]))
        for k in range(0, self._p):
            if k % 2 == 0:
                direction = list(range(0, a))
            else:
                direction = list(range(a - 2, 0, -1))
                direction.insert(0, 0)
                direction.append(a - 1)
            for i in direction:
                depth = depth + 1
                for j in range(len(first_gates_scheduled[i])):
                    if first_gates_scheduled[i][j][0] == 'Rz':
                        u = first_gates_scheduled[i][j][1][0]
                        nodes_weight = first_gates_scheduled[i][j][1][1]
                        u = {v: k for k, v in qubits_mapping_initial.items()}[u]
                        final_gates_scheduled[depth].append(
                            ('Rz', (u, 2 * gamma[k] * nodes_weight),
                             (qubits_mapping_initial[u], 2 * gamma[k] * nodes_weight)))

                    if first_gates_scheduled[i][j][0] == 'Rzz':
                        u, v = first_gates_scheduled[i][j][1][0]
                        edges_weight = first_gates_scheduled[i][j][1][1]
                        u = {v: k for k, v in qubits_mapping_initial.items()}[u]
                        v = {v: k for k, v in qubits_mapping_initial.items()}[v]
                        final_gates_scheduled[depth].append(
                            ('Rzz', ((u, v), 2 * gamma[k] * edges_weight),
                             ((qubits_mapping_initial[u], qubits_mapping_initial[v]),
                              2 * gamma[k] * edges_weight)))

                    if first_gates_scheduled[i][j][0] == 'SWAP':
                        u, v = first_gates_scheduled[i][j][1]
                        u = {v: k for k, v in qubits_mapping_initial.items()}[u]
                        v = {v: k for k, v in qubits_mapping_initial.items()}[v]
                        final_gates_scheduled[depth].append(
                            ('SWAP', (u, v),
                             (qubits_mapping_initial[u], qubits_mapping_initial[v])))
                        x = qubits_mapping_initial[v]
                        qubits_mapping_initial[v] = qubits_mapping_initial[u]
                        qubits_mapping_initial[u] = x

                    if first_gates_scheduled[i][j][0] == 'Rx':
                        u = first_gates_scheduled[i][j][1][0]
                        u = {v: k for k, v in qubits_mapping_initial.items()}[u]
                        final_gates_scheduled[depth].append(
                            ('Rx', (u, 2 * beta[k]), (qubits_mapping_initial[u], 2 * beta[k])))

        # Get the final gates scheduling sequence
        rearrange_gates_scheduled = copy.deepcopy(final_gates_scheduled)
        for i in range(len(final_gates_scheduled) - 1):
            list_bits = [i for i in range(len(final_gates_scheduled[0]))]
            if final_gates_scheduled[i]:
                if final_gates_scheduled[i][0][0] == 'SWAP':
                    list_bits = [final_gates_scheduled[i][k][1] for k in range(len(final_gates_scheduled[i]))]
                    list_bits = [n for item in list_bits for n in item]
                elif final_gates_scheduled[i][0][0] == 'Rzz':
                    list_bits = [final_gates_scheduled[i][k][1][0] for k in range(len(final_gates_scheduled[i]))]
                    list_bits = [n for item in list_bits for n in item]
                elif final_gates_scheduled[i][0][0] == 'Rz' or final_gates_scheduled[i][0][0] == 'Rx':
                    list_bits = [final_gates_scheduled[i][k][1][0] for k in range(len(final_gates_scheduled[i]))]
                else:
                    pass

            for k in range(i + 1, len(final_gates_scheduled)):
                if final_gates_scheduled[k]:
                    if final_gates_scheduled[k][0][0] == 'SWAP':
                        for j in range(len(final_gates_scheduled[k])):
                            bits = final_gates_scheduled[k][j][1]
                            if (bits[0] not in list_bits) and (bits[1] not in list_bits):
                                rearrange_gates_scheduled[i].append(final_gates_scheduled[k][j])
                                rearrange_gates_scheduled[k].remove(final_gates_scheduled[k][j])
                                list_bits.extend(bits)
                                list_bits = list(set(list_bits))
                            elif (bits[0] in list_bits) or (bits[1] in list_bits):
                                list_bits.extend(bits)
                                list_bits = list(set(list_bits))
                            else:
                                pass
                    elif final_gates_scheduled[k][0][0] == 'Rzz':
                        for j in range(len(final_gates_scheduled[k])):
                            bits = final_gates_scheduled[k][j][1][0]
                            if (bits[0] not in list_bits) and (bits[1] not in list_bits):
                                rearrange_gates_scheduled[i].append(final_gates_scheduled[k][j])
                                rearrange_gates_scheduled[k].remove(final_gates_scheduled[k][j])
                                list_bits.extend(bits)
                                list_bits = list(set(list_bits))
                            elif (bits[0] in list_bits) or (bits[1] in list_bits):
                                list_bits.extend(bits)
                                list_bits = list(set(list_bits))
                            else:
                                pass
                    elif final_gates_scheduled[k][0][0] == 'Rz' or final_gates_scheduled[k][0][0] == 'Rx':
                        for j in range(len(final_gates_scheduled[k])):
                            bits = final_gates_scheduled[k][j][1][0]
                            if bits not in list_bits:
                                rearrange_gates_scheduled[i].append(final_gates_scheduled[k][j])
                                rearrange_gates_scheduled[k].remove(final_gates_scheduled[k][j])
                                list_bits.extend((bits, bits))
                                list_bits = list(set(list_bits))
                            else:
                                pass
                    else:
                        pass
                    final_gates_scheduled = copy.deepcopy(rearrange_gates_scheduled)
                if len(list_bits) >= len(final_gates_scheduled[0]):
                    break
        k = 0
        final_gates_scheduled = defaultdict(list)
        for i in range(len(rearrange_gates_scheduled)):
            if rearrange_gates_scheduled[i]:
                final_gates_scheduled[k] = rearrange_gates_scheduled[i]
                k = k + 1

        return final_gates_scheduled, qubits_mapping_initial

    def gates_decomposition(self, final_gates_scheduled):
        # gates decomposition
        hardware_gates_scheduled = list([])
        for i in range(len(final_gates_scheduled)):
            layer_list = list([])
            layer_list1 = list([])
            layer_list2 = list([])
            for j in range(len(final_gates_scheduled[i])):
                if final_gates_scheduled[i][j][0] == 'H':
                    u = final_gates_scheduled[i][j][1]
                    layer_list.append(['H', u, 0])
                if final_gates_scheduled[i][j][0] == 'Rz':
                    u, theta = final_gates_scheduled[i][j][1]
                    layer_list.append(['Rz', u, theta])
                if final_gates_scheduled[i][j][0] == 'Rx':
                    u, theta = final_gates_scheduled[i][j][1]
                    layer_list.append(['Rx', u, theta])
                if final_gates_scheduled[i][j][0] == 'Rzz':
                    for k in range(3):
                        if k == 0:
                            u, v = final_gates_scheduled[i][j][1][0]
                            layer_list.append(['CNOT', [u, v]])
                        if k == 1:
                            u, v = final_gates_scheduled[i][j][1][0]
                            theta = final_gates_scheduled[i][j][1][1]
                            layer_list1.append(['Rz', v, theta])
                        if k == 2:
                            u, v = final_gates_scheduled[i][j][1][0]
                            layer_list2.append(['CNOT', [u, v]])
                if final_gates_scheduled[i][j][0] == 'SWAP':
                    for k in range(3):
                        if k == 0:
                            u, v = final_gates_scheduled[i][j][1]
                            ma, mi = max(u, v), min(u, v)
                            layer_list.append(['CNOT', [mi, ma]])
                        if k == 1:
                            u, v = final_gates_scheduled[i][j][1]
                            ma, mi = max(u, v), min(u, v)
                            layer_list1.append(['CNOT', [ma, mi]])
                        if k == 2:
                            u, v = final_gates_scheduled[i][j][1]
                            ma, mi = max(u, v), min(u, v)
                            layer_list2.append(['CNOT', [mi, ma]])
            if layer_list:
                hardware_gates_scheduled.append(layer_list)
            if layer_list1:
                hardware_gates_scheduled.append(layer_list1)
            if layer_list2:
                hardware_gates_scheduled.append(layer_list2)
        return hardware_gates_scheduled

    def cnot_gates_optimization(self, hardware_gates_scheduled, physical_qubits=None):
        # CNOT gates optimization
        # Two identical CNOT gates adjacent to each other are eliminated:
        # CNOT(i,j)CNOT(i,j) = identity matrix
        depth = len(hardware_gates_scheduled)
        opt_hardware_gates_scheduled = copy.deepcopy(hardware_gates_scheduled)
        for i in range(depth - 1):
            next_list = [hardware_gates_scheduled[i + 1][k][1] for k in
                         range(len(hardware_gates_scheduled[i + 1]))]
            for j in range(len(hardware_gates_scheduled[i])):
                if hardware_gates_scheduled[i][0][0] == 'CNOT':
                    if hardware_gates_scheduled[i][j][1] in next_list:
                        opt_hardware_gates_scheduled[i].remove(hardware_gates_scheduled[i][j])
                        opt_hardware_gates_scheduled[i + 1].remove(hardware_gates_scheduled[i][j])
        opt_hardware_gates_scheduled = list(filter(None, opt_hardware_gates_scheduled))

        if physical_qubits is not None:
            oq = QuantumRegister(physical_qubits, 'q')
            optimized_circuit = QuantumCircuit(oq)
            for i in range(len(opt_hardware_gates_scheduled)):
                for j in range(len(opt_hardware_gates_scheduled[i])):
                    if opt_hardware_gates_scheduled[i][j][0] == 'H':
                        optimized_circuit.h(opt_hardware_gates_scheduled[i][j][1])
                    if opt_hardware_gates_scheduled[i][j][0] == 'Rz':
                        _, v, theta = opt_hardware_gates_scheduled[i][j]
                        optimized_circuit.rz(theta, v)
                    if opt_hardware_gates_scheduled[i][j][0] == 'Rx':
                        _, v, theta = opt_hardware_gates_scheduled[i][j]
                        optimized_circuit.rx(theta, v)
                    if opt_hardware_gates_scheduled[i][j][0] == 'CNOT':
                        _, q = opt_hardware_gates_scheduled[i][j]
                        optimized_circuit.cx(q[0], q[1])
        else:
            optimized_circuit = None
        return opt_hardware_gates_scheduled, optimized_circuit

    def graph_to_qasm(self):
        qubits_mapping = self.simple_layout_mapping()
        pattern_rzz_swap, rzz_gates_cycle = self.scheduled_pattern_rzz_swap(qubits_mapping)
        best_phys2logi_mapping = self.best_initial_mapping(rzz_gates_cycle)
        print('best',best_phys2logi_mapping)
        pattern_rzz_swap_new = defaultdict(list)
        for k, v in pattern_rzz_swap.items():
            for gate in v:
                pattern_rzz_swap_new[k].append(
                    [(best_phys2logi_mapping[gate[0][0]], best_phys2logi_mapping[gate[0][1]]), gate[1]])

        final_gates_scheduled, final_phys2logi_mapping = self.QAOA_physical_circuit(pattern_rzz_swap_new,
                                                                                    best_phys2logi_mapping)
        hardware_gates_scheduled = self.gates_decomposition(final_gates_scheduled)
        opt_hardware_gates_scheduled, ScQ_circuit = self.cnot_gates_optimization(
            hardware_gates_scheduled, physical_qubits=self._physical_qubits)
        ScQ_circuit.measure_all()
        openqasm = ScQ_circuit.qasm()
        return openqasm, final_phys2logi_mapping, ScQ_circuit

    def scq_qasm(self, openqasm):
        user = User()
        user.save_apitoken(self._apitoken)
        backend = self._cloud_backend
        task = Task()
        # task.load_account()
        task.config(backend=backend)
        backend_info = task.get_backend_info()
        plt.close()
        calibration_time = backend_info['full_info']["calibration_time"]
        logical_qubits = int(re.findall(r"\d+\.?\d*", openqasm.split('qreg')[1].split(';')[0])[0])

        build_library = BuildLibrary(backend=backend, fidelity_threshold=96)
        if not os.path.exists('LibSubstructure_' + backend + '.txt'):
            print(
                "The subgraph library of " + backend + " quantum chip does not exist. Please wait: creating subgraph library!")
            substructure_data = build_library.build_substructure_library()
            print('Complete building!')
        else:
            with open('LibSubstructure_' + backend + '.txt', 'r', encoding='utf-8') as f:
                substructure_data = eval(f.read())
                if substructure_data['calibration_time'] != calibration_time:
                    print(
                        "The qubits of " + backend + " quantum chip have been recalibrated. Please wait: updating the subgraph library of the corresponding quantum chip!")
                    build_library.build_substructure_library()
                    print('Complete building!')
                else:
                    print(
                        "The information of qubits are unchanged, and the existing subgraph library is directly called!")

        physical_qubits = len(substructure_data['qubit_to_int'])
        scq_qasm = openqasm.replace('qreg q[' + str(logical_qubits),
                                    'qreg q[' + str(physical_qubits))

        if backend == "ScQ-P136":
            scq_qasm = re.sub(r'barrier.*\n', "", scq_qasm)
            if not os.path.exists('LibSubchain_' + backend + '.txt'):
                print("The one-dimensional chain library of " + backend + " quantum chip does not exist, and is being created!")
                chain_data = build_library.chain_library_2D(substructure_data)
                print('Complete building!')
            else:
                with open('LibSubchain_' + backend + '.txt', 'r', encoding='utf-8') as f:
                    chain_data = eval(f.read())
                    if chain_data['calibration_time'] != calibration_time:
                        print("Waiting: Building a library of one-dimensional chain structures for the " + backend + " quantum chip!")
                        chain_data = build_library.chain_library_2D(substructure_data)
                        print('Complete building!')

            longset_chain = max(chain_data['subchain_dict'].keys())
            if logical_qubits > longset_chain:
                raise SystemExit(
                    "Currently, " + backend + " quantum chip supports a maximum of " + str(longset_chain) + " qubits!")

            best_chain = chain_data['subchain_dict'][logical_qubits][0]
            nodes = []
            edges = []
            for k in best_chain:
                if k[0] not in nodes:
                    nodes.append(k[0])
                if k[1] not in nodes:
                    nodes.append(k[1])
                edges.append(k[0:2])

            G2 = nx.DiGraph()
            G2.add_nodes_from(nodes)
            G2.add_edges_from(edges)

            node_degree = dict(G2.degree)
            sort_degree = sorted(node_degree.items(), key=lambda kv: kv[1], reverse=False)
            qubits_list = []
            begin_node = sort_degree[0][0]
            qubits_list.append(begin_node)
            for k in range(len(sort_degree)):
                neighbor_nodes = [k for k, v in G2[begin_node].items()]
                neighbor_nodes = [node for node in neighbor_nodes if node not in qubits_list]
                if neighbor_nodes:
                    qubits_list.append(neighbor_nodes[0])
                    begin_node = neighbor_nodes[0]
                else:
                    break

        else:
            longset_chain = max(substructure_data['substructure_dict'].keys())
            if logical_qubits > longset_chain:
                raise SystemExit(
                    "Currently, " + backend + " quantum chip supports a maximum of " + str(longset_chain) + " qubits!")
            sub_clist = [qubits[0:2] for qubits in substructure_data['substructure_dict'][logical_qubits][0]]
            qubits_list = []
            for coupling in sub_clist:
                if coupling[0] not in qubits_list:
                    qubits_list.append(coupling[0])
                if coupling[1] not in qubits_list:
                    qubits_list.append(coupling[1])
            qubits_list = sorted(qubits_list)

        print('physical qubits used:\n', qubits_list)

        req_to_q = {q: qubits_list[q] for q in range(len(qubits_list))}
        for req, q in sorted(req_to_q.items(), key=lambda item: item[1], reverse=True):
            scq_qasm = scq_qasm.replace('q[' + str(req) + ']', 'q[' + str(q) + ']')
        scq_qasm = scq_qasm.replace('meas[', 'c[')
        plt.close()
        return scq_qasm

    def send(self, wait=True, shots=1024, task_name: str = ''):
        """
        Send the task to the quafu cloud platform.
        Args:
            wait (bool): If you choose wait=Ture, you have to wait for the result
                         to return, during which the program will not terminate.
                         If you choose wait=False, you can submit tasks asynchronously
                         without waiting online for the results to return.
            shots (int): The number of sampling of quantum computer.
            task_name (str): The name of the task so that you can query the task status
                             on the quafu cloud platform later.
        Returns:
            task_id (str): The ID number of the task, which can uniquely identify the task.
        """
        openqasm, final_phys2logi_mapping, _ = self.graph_to_qasm()
        self._logi2phys_mapping = {v: k for k, v in final_phys2logi_mapping.items()}
        scq_qasm = self.scq_qasm(openqasm)
        # send to quafu cloud
        qubits = int(re.findall(r"\d+\.?\d*", scq_qasm.split('qreg')[1].split(';')[0])[0])
        q = quafuQC(qubits)
        q.from_openqasm(scq_qasm)
        task = Task()
        # task.load_account()
        task.config(backend=self._cloud_backend, shots=shots, compile=False)
        task_id = task.send(q, wait=wait, name=task_name, group=task_name).taskid
        print("The task has been submitted to the quafu cloud platform.\nThe task ID is '%s'" % task_id)
        return task_id

    def task_status_query(self, task_id: str):
        task = Task()
        # task.load_account()
        task_status = task.retrieve(task_id).task_status
        if task_status=='In Queue' or task_status=='Running':
            print("The current task status is '%s', please wait." % task_status)
        elif task_status=='Completed':
            print("The task execution has completed and the result has been returned.")
            res = task.retrieve(task_id)
            return res

    def right_left_counts_rearrange(self, logi2phys_mapping, counts):
        # The bits are arranged as 0, 1, 2,... from right to the left.
        qubit_str = list(counts.keys())
        counts_new = defaultdict(list)
        for i in range(len(qubit_str)):
            str = ''
            for k in range(len(logi2phys_mapping)):
                str = str + qubit_str[i][::-1][logi2phys_mapping[k]]
            counts_new[str[::-1]] = counts[qubit_str[i]]
        return counts_new

    def left_right_counts_rearrange(self, logi2phys_mapping, counts):
        # The bits are arranged as 0, 1, 2,... from left to the right.
        qubit_str = list(counts.keys())
        counts_new = defaultdict(list)
        for i in range(len(qubit_str)):
            str = ''
            for k in range(len(logi2phys_mapping)):
                str = str + qubit_str[i][::][logi2phys_mapping[k]]
            counts_new[str] = counts[qubit_str[i]]
        return counts_new

    def graph_sampling_energy_ising(self, counts):
        # Obtain QAOA energy from circuit sampling results
        counts_energy = {}
        for i in range(len(counts)):
            # 0 -> -1, 1 -> 1
            result1 = [int(u) for u in list(counts[i][0])]
            # result1.reverse()
            energy1 = 0
            for node in self._nodes:
                energy1 = energy1 + node[1] * (2 * result1[node[0]] - 1)
            for edge in self._edges:
                energy1 = energy1 + 2* edge[2] * (2 * result1[edge[0]] - 1) * (2 * result1[edge[1]] - 1)
            counts_energy[counts[i]] = energy1
        counts_energy = sorted(counts_energy.items(), key=lambda x: x[1])
        return counts_energy

    def graph_sampling_energy_qubo(self, counts):
        # Obtain QUBO cost from circuit sampling results
        counts_energy = {}
        import numpy as np
        qubo_mat = np.zeros([len(self._nodes), len(self._nodes)])
        ising_mat = np.zeros([len(self._nodes), len(self._nodes)])
        for edge in self._edges:
            ising_mat[edge[0], edge[1]] = edge[2]
            ising_mat[edge[1], edge[0]] = edge[2]
            qubo_mat[edge[0], edge[1]] = 4*edge[2]/2.
            qubo_mat[edge[1], edge[0]] = 4*edge[2]/2.
        for node in self._nodes:
            ising_mat[node[0],node[0]] = node[1]
            qubo_mat[node[0],node[0]] = (node[1]-sum(2.*qubo_mat[node[0]]/4.))*2
        for i in range(len(counts)):
            result1 = np.array([int(u) for u in list(counts[i][0])])
            energy1 = np.dot(np.dot(result1, qubo_mat), result1)
            counts_energy[counts[i]] = energy1
        counts_energy = sorted(counts_energy.items(), key=lambda x: x[1])
        # print('qubo_mat',qubo_mat)
        return counts_energy

    def results_processing(self, results):
        counts_ScQ0 = results.res
        logi2phys_mapping = self._logi2phys_mapping
        counts_ScQ_new = self.left_right_counts_rearrange(logi2phys_mapping, counts_ScQ0)
        counts_ScQ = sorted(counts_ScQ_new.items(), key=lambda x: x[1], reverse=True)
        counts_ScQ = [(item[0], item[1]) for item in counts_ScQ]

        for elem in counts_ScQ:
            if elem[1] == 0:
                counts_ScQ.remove(elem)

        counts_energy = self.graph_sampling_energy_qubo(counts_ScQ)
        print('Results ((qubits str, number of sampling), QUBO Cost):\n', counts_energy)
        return counts_energy

    def visualization(self, counts_energy, problem='MaxCut', solutions=3, problem_graph=None):
        # draw result
        plt.close()
        print('Send results to client:')
        if problem=='MaxCut':
            for s in range(solutions):
                optimal_solution = counts_energy[s][0][0]
                cut_node1 = []
                cut_node2 = []
                for i in range(len(optimal_solution)):
                    if optimal_solution[i] == '0':
                        cut_node1.append(i)
                    else:
                        cut_node2.append(i)
                pos = nx.spring_layout(self._g)
                # pos = nx.circular_layout(self._g)
                nx.draw_networkx(self._g, pos=pos, nodelist=cut_node1, node_size=500, node_color='c', font_size=15, width=2)
                nx.draw_networkx(self._g, pos=pos, nodelist=cut_node2, node_size=500, node_color='r', font_size=15, width=2)
                nx.draw_networkx_edges(self._g, pos, width=2, edge_color='g', arrows=False)
                # plt.axis('off')
                plt.tight_layout()
                plt.show()
                print('======== Solution ' + str(1+s) + ' ========')
                print('Cluster 1:', sorted(cut_node1))
                print('Cluster 2:', sorted(cut_node2))
                print('Cost (QUBO):', counts_energy[s][1])
        elif problem=='GraphColoring':
            import matplotlib.colors
            color_list = ['r', 'c', 'b', 'y', 'm']
            color_list = color_list + list(matplotlib.colors.cnames.values())
            nodes = len(problem_graph.nodes)
            color_num = int(len(counts_energy[0][0][0])/nodes)
            for s in range(solutions):
                print('======== Solution ' + str(1 + s) + ' ========')
                optimal_solution = counts_energy[s][0][0]
                qubo_bits = []
                for n in range(0, nodes):
                    qubo_bits.append(optimal_solution[n*color_num:(n+1)*color_num])
                color_dic = {}
                for k in range(nodes):
                    color_dic.setdefault(qubo_bits[k], []).append(k)
                pos = nx.spring_layout(problem_graph)
                # pos = nx.circular_layout(self._g)
                color = 0
                for bit, node_list in color_dic.items():
                    nx.draw_networkx(problem_graph, pos=pos, nodelist=node_list, node_size=500, node_color=color_list[color], font_size=15, width=2)
                    color = color + 1
                    print('Coloring ' + str(color)+ ':', sorted(node_list))
                nx.draw_networkx_edges(problem_graph, pos, width=2, edge_color='g', arrows=False)
                # plt.axis('off')
                plt.tight_layout()
                plt.show()
                print('Cost (QUBO):', counts_energy[s][1])
        else:
            print('The current version does not support visualization of this type of problem!')



# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import logging
import time
import numpy as np
import networkx as nx
import random
from numpy.random import choice
import matplotlib.pyplot as plt
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_regular_graph


logger = logging.getLogger(__name__)


class MinimumVertexCover:
    """
    MVC (Minimum Vertex Cover) problem:
    For an undirected graph, find a vertex cover (a subset of the vertices/nodes) 
    with a minimum number of vertices in the subset
    such that each edge in the graph is incident to at least one vertex in the subset.
    For the weight graph, the weight sum of vertices in the subset is minimum.
    """
    def __init__(self,
                 graph: nx.Graph = None,
                 node_num: int = None,
                 node_degree: int = 3,
                 weight_range: int = 10,
                 P: float = None,
                 seed: int = None):
        """
        Args:
            graph (nx.Graph): an networkx graph generated from input adjacency matrix 
            node_num (int): number of nodes in the graph
            node_degree (int): node degree of the graph, default value is 3
            P (int): the penalty value for the penalty terms
            (require input: adjacency matrix, P)
            
         Returns:
            node_num, graph, P
            
        Example:
             adjacency_matrix = np.array([[0, 1, 1, 0, 0],
                              [1, 0, 0, 1, 0],
                              [1, 0, 0, 1, 1],
                              [0, 1, 1, 0, 1],
                              [0, 0, 1, 1, 0]])
             g = nx.from_numpy_matrix(adjacency_matrix)
             penalty = 8
        """
        
        if graph is None:
            # generate random graph according to node_num and node_degree and weight_range
            self._node_num = node_num
            self._graph = random_regular_graph(node_num=self._node_num,
                                               degree=node_degree,
                                               weight_range=weight_range,
                                               seed=seed)
            
            self._P = P
        else:
            self._node_num = len(graph.nodes)
            self._graph = graph
            self._P = P

        self._qmatrix = None
        self._shift = None
        
    @property
    def node_num(self):
        return self._node_num

    @property
    def graph(self):
        return self._graph
    @property
    def weight_range(self):
        return self._weight_range

    def update_random_graph(self, node_num, node_degree, weight_range, seed):
        self._node_num = node_num
        self._graph = random_regular_graph(node_num=self._node_num,
                                           degree=node_degree,
                                           weight_range=weight_range,
                                        seed=seed)

    def get_Qmatrix(self):
        """
        get the Q matrix in QUBO model of MVC problem
        Args:
            adjacent_mat (np.array): the adjacent matrix of graph G of the problem

        Returns:
            q_mat (np.array): the the Q matrix of QUBO model.
        
        ..math::
            minimise x(T)Qx
            Q[i][j] = P/2 
            Q[i][i] = -weight[i] - P * node degree[i]
        """
        adj_mat = nx.adjacency_matrix(self._graph).A
        qubo_mat = np.array(adj_mat, dtype='float64')
        
        for i in range(self._node_num):
            qubo_mat[i][i] = 1 - self._P * self._graph.degree[i] #node degree
            for j in range(self._node_num):
                if i == j:
                    continue
                elif abs(adj_mat[i][j] - 0.) <= 1e-8:
                    qubo_mat[i][j] = 0.0
                else:
                    qubo_mat[i][j] = self._P / 2.0
                    
        shift = self._P
        
        return qubo_mat, shift

    def minimum_vertex_cover_value(self, x, w):
        """Compute the value of a cut.

        Args:
            x (numpy.ndarray): binary string as numpy array.
            w (numpy.ndarray): adjacency matrix.

        Returns:
            float: value of the cut.
        """
        if self._qmatrix is None:
            self._qmatrix = self.get_Qmatrix()

        X = np.matmul(x, np.matmul(self._qmatrix, np.transpose(x))) 
        return X
    
    def run(self):
        if self._qmatrix is None:
            self._qmatrix, self._shift = self.get_Qmatrix()

        qubo_mat = self._qmatrix
        ising_mat = get_ising_matrix(qubo_mat)
        mvc_graph = get_weights_graph(ising_mat)
        return mvc_graph, self._shift


# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import os
import sys
import time
import logging
import random
import numpy as np
import networkx as nx
import matplotlib.pyplot as plt
from Qcover.applications.common import get_ising_matrix, get_weights_graph, random_regular_graph


logger = logging.getLogger(__name__)


class SherringtonKirkpatrick:
    """
    Sherrington Kirkpatrick problem: The Sherrington-Kirkpatrick model is a special instance of a spin glass.
    it's defined on the complete graph with wij randomly chosen to be ±1
    """
    def __init__(self, node_num: int = 5):
        self._node_num = node_num

    @property
    def node_num(self):
        return self._node_num

    def run(self):
        ising_mat = np.zeros((self._node_num, self._node_num)) #, dtype=float
        for i in range(self._node_num):
            for j in range(self._node_num):
                if i <= j:
                    ising_mat[i][j] = np.random.choice([1, -1], 1)
                else:
                    ising_mat[i][j] = ising_mat[j][i]
        sk_graph = get_weights_graph(ising_mat)
        return sk_graph



# This code is part of Qcover.
#
# (C) Copyright BAQIS 2021, 2022.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Object to solve QAOA problems

The QAOA problems can be represented as an Ising model, and then be transformed to a DAG.
The directed acyclic graph is decomposed by a specified p value, and these subgraphs then
be transformed as circuits and be executed on simulators, using optimizer to get the
optimal parameters of the original Ising model
"""

import sys
import time
from itertools import permutations

from typing import Optional
from collections import defaultdict
import numpy as np
import networkx as nx
from Qcover.utils import get_graph_weights, generate_weighted_graph
from Qcover.optimizers import Optimizer, COBYLA
from Qcover.backends import Backend, CircuitByQiskit, CircuitByTensor
from Qcover.exceptions import GraphTypeError, UserConfigError
import warnings
warnings.filterwarnings("ignore")


class Qcover:
    """
    Qcover is a QAOA solver
    """
    # pylint: disable=invalid-name
    def __init__(self,
                 graph: nx.Graph = None,
                 p: int = 1,
                 optimizer: Optional[Optimizer] = COBYLA(),
                 backend: Optional[Backend] = CircuitByQiskit(),
                 research_obj: str = "QAOA"
                 ) -> None:

        assert graph is not None
        self._simple_graph = graph
        self._p = p
        self._research_obj = research_obj
        self._backend = backend
        self._backend._p = p
        self._backend._origin_graph = graph
        self._backend._research = research_obj
        self._optimizer = optimizer
        self._optimizer._p = p

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, new_p):
        self._p = new_p

    @property
    def backend(self):
        return self._backend

    @backend.setter
    def backend(self, new_backend):
        self._backend = new_backend

    @property
    def optimizer(self):
        return self._optimizer

    @optimizer.setter
    def optimizer(self, new_optimizer):
        self._optimizer = new_optimizer

    # def generate_subgraph(self, dtype: str):
    def graph_decomposition(self, p):
        """
        according to the arguments of dtype and p to generate subgraphs from graph
        Args:
            graph (nx.Graph): graph to be decomposed
            dtype (string): set "node" or "edge", the ways according to which to decompose the graph
            p (int): the p of subgraphs
        Return:
            subg_dict (dict) form as {node_id : subg, ..., (node_id1, node_id2) : subg, ...}
        """
        # if dtype not in ["node", "edge"]:
        #     print("Error: wrong dtype, dtype should be node or edge")
        #     return None
        if p <= 0:
            warnings.warn(" the argument of p should be >= 1 in qaoa problem, "
                          "so p would be set to the default value at 1")
            p = 1

        nodes_weight, edges_weight = get_graph_weights(self._simple_graph)

        subg_dict = defaultdict(list)
        # if dtype == 'node':
        for node in self._simple_graph.nodes:
            node_set = {(node, nodes_weight[node])}
            edge_set = set()
            for i in range(p):
                new_nodes = {(nd2, nodes_weight[nd2]) for nd1 in node_set for nd2 in self._simple_graph[nd1[0]]}
                new_edges = {(nd1[0], nd2, edges_weight[nd1[0], nd2]) for nd1 in node_set for nd2 in self._simple_graph[nd1[0]]}
                node_set |= new_nodes
                edge_set |= new_edges

            subg = generate_weighted_graph(node_set, edge_set)
            subg_dict[node] = subg
        # else:
        for edge in self._simple_graph.edges:
            node_set = {(edge[0], nodes_weight[edge[0]]), (edge[1], nodes_weight[edge[1]])}
            edge_set = {(edge[0], edge[1], edges_weight[edge[0], edge[1]])}

            for i in range(p):
                new_nodes = {(nd2, nodes_weight[nd2]) for nd1 in node_set for nd2 in self._simple_graph[nd1[0]]}
                new_edges = {(nd1[0], nd2, edges_weight[nd1[0], nd2]) for nd1 in node_set for nd2 in
                             self._simple_graph.adj[nd1[0]]}
                node_set |= new_nodes
                edge_set |= new_edges

            subg = generate_weighted_graph(node_set, edge_set)
            subg_dict[edge] = subg
        return subg_dict

    def calculate(self, pargs, p=None):
        """
        The framework function which use the backend to calculate the value of expectation,
        and be used as the object function in the optimization function of the optimizer
        Args:
            pargs: the value of the parameter alpha and beta in the circuit
            p: the integer used to define the number of layers the current circuit needs to be superimposed
        Returns:
            the value of expectation calculated by backends
        """
        p = self._p if p is None else p
        element_to_graph = self.graph_decomposition(p=p)

        # checking graph type of given problem
        if not isinstance(self._backend, CircuitByTensor):
            for k, v in element_to_graph.items():
                ncnt, ecnt = len(v.nodes), len(v.edges)
                try:
                    nreq1 = ncnt * (ncnt - 1) <= 2 * ecnt and ncnt >= 20
                    nreq2 = isinstance(k, int) and v.degree[k] >= 30
                    if nreq1 or nreq2:
                        raise GraphTypeError("The problem is transformed into a dense graph, " \
                           "which is difficult to be solved effectively by Qcover")
                except GraphTypeError as e:
                    print(e)
                    # if not self._hard_to_calcute:
                    #     print(e)
                    #     self._hard_to_calcute = True
                    # sys.exit()

        self._backend._pargs = pargs
        self._backend._element_to_graph = element_to_graph
        return self._backend.expectation_calculation(p)

    def run(self, is_parallel=False):
        self._backend._is_parallel = is_parallel
        x, fun, nfev = self._optimizer.optimize(objective_function=self.calculate)
        res = {"Optimal parameter value": x, "Expectation of Hamiltonian": fun, "Total iterations": nfev}
        return res

# usage example
if __name__ == '__main__':

    p = 1
    g = nx.Graph()
    nodes = [(0, 0), (1, 0), (2, 0)]
    edges = [(0, 1, 1), (1, 2, 1)]

    # nodes = [(0, 0), (1, 0), (2, 0), (3, 0), (4, 0)]
    # edges = [(0, 1, -1), (1, 2, -1), (2, 3, -1), (3, 4, -1)]

    for nd in nodes:
        u, w = nd[0], nd[1]
        g.add_node(int(u), weight=int(w))
    for ed in edges:
        u, v, w = ed[0], ed[1], ed[2]
        g.add_edge(int(u), int(v), weight=int(w))

    # node_num = 10
    # for nd in range(node_num):
    #     g.add_node(nd, weight=0)
    #     if nd < node_num - 1:
    #         g.add_edge(nd, nd + 1, weight=-1)

    #Ising test
    # g.add_node(0, weight=0)
    # for i in range(1, 30):
    #     g.add_node(i, weight=0)
    #     g.add_edge(i, i - 1, weight=-1)

    # nx.draw(g)
    # from Qcover.applications import MaxCut
    # mxt = MaxCut(g)
    # mxt = MaxCut(node_num=20, node_degree=3)
    #10 3 0.035  100 3 0.029
    #10 6 0.028  100 6 134.56/ce
    # g, shift = mxt.run()

    from Qcover.optimizers import GradientDescent, SPSA, Interp, SHGO

    optg = GradientDescent(options={'maxiter':300, 'learning_rate':0.05, 'tol':1e-6})
    opti = Interp(optimize_method="COBYLA", options={'tol': 1e-8, 'disp': False})
    opta = SPSA(options={'maxiter':300, 'tol':1e-6})
    optc = COBYLA(options={'tol': 1e-3, 'disp': True})
    opts = SHGO(options={'minimizer_kwargs':{'method':'SLSQP', 'options':{'ftol': 1e-12}}, 'sampling_method':'sobol'})
    qiskit_bc = CircuitByQiskit(expectation_calc_method="statevector")
    from Qcover.backends import CircuitByQton, CircuitByQulacs, CircuitByProjectq, CircuitByTensor, CircuitByCirq
    qulacs_bc = CircuitByQulacs()
    # cirq_bc = CircuitByCirq()
    # projectq_bc = CircuitByProjectq()
    # ts = CircuitByTensor()
    # qt = CircuitByQton()  #expectation_calc_method="tensor"

    qc = Qcover(g, p,
                # research_obj="QAOA",
                optimizer=optc,  #@ optc,
                backend=qulacs_bc)  #qiskit_bc, qt, , cirq_bc, projectq_bc

    st = time.time()
    sol = qc.run(is_parallel=False)  #True
    ed = time.time()
    print("time cost by QAOA is:", ed - st)
    print("solution is:", sol)
    params = sol["Optimal parameter value"]
    qc.backend._pargs = params
    out_count = qc.backend.get_result_counts(params)
    res_exp = qc.backend.expectation_calculation()
    print("the optimal expectation is: ", res_exp)
    qc.backend.sampling_visualization(out_count)

    # test RQAOA
    # import matplotlib.pyplot as plt
    # from Qcover.applications import MaxCut
    #
    # p = 1
    # node_num = [10, 50, 100, 500, 1000]
    # node_d = [3, 4, 5, 6]
    # for i in node_num:
    #     et, gt = [], []
    #     ev, gv = [], []
    #     for nd in node_d:
    #         print("node number is %d, node degree is %d" % (i, nd))
    #         print("----------------------------------------------")
    #         mxt = MaxCut(node_num=i, node_degree=nd)
    #         g, shift = mxt.run()
    #
    #         g_exp = g.copy()
    #         qc_exp = Qcover(g_exp, p,
    #                     optimizer=COBYLA(options={'tol': 1e-3, 'disp': True}),
    #                     backend=CircuitByQiskit(expectation_calc_method="statevector"))  # qt, ,qulacs_bc
    #
    #         st = time.time()
    #         sol_exp, sexp_exp = qc_exp.run(mode='RQAQA', node_threshold=1, iter_time=3, corr_method='expectation')
    #         ed = time.time()
    #         t1 = ed - st
    #
    #         g_g = g.copy()
    #         qc_g = Qcover(g_g, p,
    #                     optimizer=COBYLA(options={'tol': 1e-3, 'disp': True}),
    #                     backend=CircuitByQiskit(expectation_calc_method="statevector"))  # qt, ,qulacs_bc
    #         st = time.time()
    #         sol_g, sexp_g = qc_g.run(mode='RQAQA', node_threshold=1, iter_time=3, corr_method='g')
    #         ed = time.time()
    #         t2 = ed - st
    #
    #         exp_e, exp_g = 0, 0
    #         for (x, y) in g.nodes.data('weight', default=0):
    #             exp_e += y * (sol_exp[x] * 2 - 1)
    #             exp_g += y * (sol_g[x] * 2 - 1)
    #         for (u, v, c) in g.edges.data('weight', default=0):
    #             exp_e += c * (sol_exp[u] * 2 - 1) * (sol_exp[v] * 2 - 1)
    #             exp_g += c * (sol_g[u] * 2 - 1) * (sol_g[v] * 2 - 1)
    #
    #         et.append(t1)
    #         gt.append(t2)
    #
    #         ev.append(exp_e)
    #         gv.append(exp_g)
    #
    #         plt.figure(0)
    #         plt.plot(range(len(sexp_exp)), sexp_exp, "ob-", label="origin")
    #         plt.plot(range(len(sexp_g)), sexp_g, "^r-", label="correlation")
    #         plt.ylabel('expectation')
    #         plt.xlabel('iteration rounds')
    #         plt.title("node is %d, degree is %d" % (i, nd))
    #         plt.legend()
            # plt.savefig('/home/baqis/Qcover/result_log/correlation_exp/iter_%d_%d.png' % (i, nd))
            # plt.savefig('E:/Working_projects/QAOA/Qcover/result_log/correlation_exp/iter_%d_%d.png' % (i, nd))
        #     plt.savefig('/home/puyanan/Qcover/result_log/correlation_exp/iter_%d_%d.png' % (i, nd))
        #     plt.close('all')
        #
        # plt.figure(1)
        # plt.plot(node_d, et, "ob-", label="origin")
        # plt.plot(node_d, gt, "^r-", label="correlation")
        # plt.ylabel('Time cost')
        # plt.xlabel('node degree')
        # plt.title("node is %d" % i)
        # plt.legend()
        # plt.savefig('/home/baqis/Qcover/result_log/correlation_exp/time_cost_%d.png' % i)
        # plt.savefig('E:/Working_projects/QAOA/Qcover/result_log/correlation_exp/time_cost_%d.png' % i)
        # plt.savefig('/home/puyanan/Qcover/result_log/correlation_exp/time_cost_%d.png' % i)
        #
        # plt.figure(2)
        # plt.plot(node_d, ev, "*g-", label="origin")
        # plt.plot(node_d, gv, "dy-", label="correlation")
        # plt.ylabel('expectation of hamitonian')
        # plt.xlabel('node degree')
        # plt.title("node is %d" % i)
        # plt.legend()
        # plt.savefig('/home/baqis/Qcover/result_log/correlation_exp/hamitonian_exp_%d.png' % i)
        # plt.savefig('E:/Working_projects/QAOA/Qcover/result_log/correlation_exp/hamitonian_exp_%d.png' % i)
        # plt.savefig('/home/puyanan/Qcover/result_log/correlation_exp/hamitonian_exp_%d.png' % i)
        #
        # plt.close('all')




# print("time cost by RQAOA with expectation is:", t1)
# print("expectation list with expectation:", exp_e)
# print("solution of expectation is:", sol_exp)
# # {8: 0, 4: 1, 2: 0, 3: 1, 7: 0, 5: 1, 0: 0, 6: 0, 1: 1, 9: 1})
#
# print("time cost by RQAOA with correlation is:", t2)
# print("expectation list with correlation:", exp_g)
# print("solution of correlation is:", sol_g)







# This file should be run by qton, which can be got at the URL below:
# https://github.com/thewateriswide/qton_2.1

from Qcover.core import Qcover

from typing import Optional

from Qcover.backends import Backend,CircuitByQulacs

from Qcover.simulator import Qcircuit

from Qcover.optimizers import Optimizer,COBYLA

from Qcover.utils import get_graph_weights, generate_weighted_graph, generate_graph_data

import networkx as nx

import numpy as np

import warnings
warnings.filterwarnings("ignore")

class QAOA_Ent_TMI:
    
    def __init__(self,
                 p: int = 1,
                 graph: nx.Graph = None,
                 optimizer: Optional[Optimizer] = COBYLA(),
                 backend: Optional[Backend] = CircuitByQulacs(),
                 ) -> None:
        
        assert graph is not None
        self._p = p
        self._original_graph = graph

        self._qc = Qcover(self._original_graph,
                          self._p,
                          optimizer=optimizer,
                          backend=backend,
                          research_obj="QAOA")

    @property
    def p(self):
        return self._p

    @p.setter
    def p(self, ap):
        self._p = ap

    @property
    def qc(self):
        return self._qc

    @property
    def original_graph(self):
        return self._original_graph

    @original_graph.setter
    def original_graph(self, graph):
        """
        according to the type of graph(nx.graph / tuple) to set the value of
        self._original_graph
        """
        if isinstance(graph, nx.Graph):
            self._original_graph = graph
        elif isinstance(graph, tuple):
            assert (len(graph) >= 2) and (len(graph) <= 3)

            if len(graph) == 2:
                node_num, edge_num = graph
                wr = None
            elif len(graph) == 3:
                node_num, edge_num, wr = graph

            nodes, edges = generate_graph_data(node_num, edge_num, wr)
            self._original_graph = generate_weighted_graph(nodes, edges)
        elif isinstance(graph, list):
            assert len(graph) == 3
            node_list, edge_list, weight_range = graph
            self._original_graph = generate_weighted_graph(node_list, edge_list, weight_range)
        else:
            print("Error: the argument graph should be a instance of nx.Graph "
                  "or a tuple formed as (node_num, edge_num)")


    def run(self, is_parallel=False):
        res = self._qc.run(is_parallel=is_parallel)  # True
        return res

    def QAOA_gate(self,qc, G, theta):
        
        p = len(theta)//2
        gamma = theta[:p]
        beta = theta[p:]
    
        #N qubits#
        for j in list(G.nodes()):
            qc.h(j)
    
        ### U transformation ###
        for k in range(p):
            for i in list(G.nodes()):
                qc.rz(2 * gamma[k]*nx.get_node_attributes(G, "weight")[i],i)
        
            for pair in list(G.edges()):
                qc.rzz(2*gamma[k]*nx.get_edge_attributes(G, "weight")[(pair[0],pair[1])], pair[0], pair[1])
        
            for nd in list(G.nodes()):
                qc.rx(2*beta[k], nd)
                
        return None

    def von_entropy(rho):
        e, v= np.linalg.eig(rho)
        S=-(e*np.log(e)/np.log(2)).sum().real
        return S

    def Ent_TMI_Channel(self, G, list_C, theta):
        Qubit_Num=len(list(G.nodes()))
        ##########one-qubit entanglement
        a=[]
        b=[]
        Smean01=[]
        Num_list=[]
        for n in range(Qubit_Num):
            Num_list.append(n)
        ############entanglement entropy per layer
        for i in range(len(theta)//2):  #int()
            a.append(theta[i])
            b.append(theta[int(len(theta)/2)+i])
            cta=a+b
            Sc01=[]
            for j in range(Qubit_Num):
                fin_circ1=Qcircuit(Qubit_Num,backend='density_matrix')
                self.QAOA_gate(fin_circ1, G, cta)
                fin_circ1.reduce(Num_list[j])
                rho_c1=fin_circ1.state.copy()
                S_01=self.von_entropy(rho_c1)
                Sc01.append(S_01)
        
            smean01=np.mean(Sc01)
            Smean01.append(smean01)
        

        #######qton calculate I3
        I3_list=[]
        qc=Qcircuit(Qubit_Num,backend='unitary')
        qaoag=self.QAOA_gate(qc,len(theta)/2, G, theta)
        u=qc.state.copy()
        v=qc.state.reshape(2^(Qubit_Num)*2^(Qubit_Num))/np.sqrt(2^(Qubit_Num))
        rho=np.outer(v,v.conjugate())

        ########ABCD systems
        list_A=[0]
        list_B=list(set(list(G.nodes()))-set(list_A))
        list_pair=[]
        for i in list(G.nodes()):
            list_pair.append(i+Qubit_Num)
        list_D=list(set(list_pair)-set(list_C))

        qu_ac=Qcircuit(2*Qubit_Num,backend='density_matrix')
        qu_ac.state=rho
        qu_ac.reduce(list_B+list_D)
        S_ac=self.von_entropy(qu_ac.state)

        qu_ad=Qcircuit(2*Qubit_Num,backend='density_matrix')
        qu_ad.state=rho
        qu_ad.reduce(list_B+list_C)
        S_ad=self.von_entropy(qu_ad.state)

        qu_cd=Qcircuit(2*Qubit_Num,backend='density_matrix')
        qu_cd.state=rho
        qu_cd.reduce(list_A+list_B)
        S_cd=self.von_entropy(qu_cd.state)

        qu_a=Qcircuit(2*Qubit_Num,backend='density_matrix')
        qu_a.state=rho
        qu_a.reduce(list_B+list_C+list_D)
        S_a=self.von_entropy(qu_a.state)

        qu_c=Qcircuit(2*Qubit_Num,backend='density_matrix')
        qu_c.state=rho
        qu_c.reduce(list_A+list_B+list_D)
        S_c=self.von_entropy(qu_c.state)

        qu_d=Qcircuit(2*Qubit_Num,backend='density_matrix')
        qu_d.state=rho
        qu_d.reduce(list_A+list_B+list_C)
        S_d=self.von_entropy(qu_d.state)

        qu_acd=Qcircuit(2*Qubit_Num,backend='density_matrix')
        qu_acd.state=rho
        qu_acd.reduce(list_B)
        S_acd=self.von_entropy(qu_acd.state)


        I3=S_a+S_c+S_d-S_cd-S_ac-S_ad+S_acd
            
        I3_list.append(I3)

        return I3_list,Smean01

if __name__ == '__main__':
    p = 5
    g = nx.Graph()
    nodes = [(0, 1), (1, 1), (2, 1),(3, 1),(4, 1),(5, 1),(6, 1)]
    edges = [(0, 1, -1), (1, 2, -1),(2, 3, -1),(3, 4, -1),(4, 5, -1),(5, 6, -1)]

    for nd in nodes:
        u, w = nd[0], nd[1]
        g.add_node(int(u), weight=int(w))
    for ed in edges:
        u, v, w = ed[0], ed[1], ed[2]
        g.add_edge(int(u), int(v), weight=int(w))

    qcover_bc = CircuitByQulacs()
    optc = COBYLA(options={'disp': True,'tol': 1e-3, 'maxiter': 10000})  #

    qaoa_ent_tmi = QAOA_Ent_TMI(graph=g,
                           p=p,
                           optimizer=optc,
                           backend=qcover_bc) 

    res = qaoa_ent_tmi.run()

    print("solution is:", res)
    params = res["Optimal parameter value"]
    list_C=[13,12,11]

    channel_res=qaoa_ent_tmi.Ent_TMI_Channel(g,list_C,params)

    print("entanglement entropy at every layer:",channel_res[1])
    print("tripartite information at every layer:",channel_res[0])

    


from .backend import Backend
from .circuitbyqiskit import CircuitByQiskit
from .circuitbyprojectq import CircuitByProjectq
from .circuitbycirq import CircuitByCirq
from .circuitbyqulacs import CircuitByQulacs
# from .circuitbytket import CircuitByTket
from .circuitbytensor import CircuitByTensor
from .circuitbyqton import CircuitByQton
import warnings
warnings.filterwarnings("ignore")

__all__ = [
    'Backend',
    'CircuitByCirq',
    'CircuitByQiskit',
    'CircuitByProjectq',
    'CircuitByTensor',
    'CircuitByQulacs',
    'CircuitByQton'
]



import networkx as nx
import numpy as np

def get_graph_weights(graph):
    """
    get the weights of nodes and edges in graph
    Args:
        graph (nx.Graph): graph to get weight of nodes and edges
    Return:
        node weights form is dict{nid1: node_weight}, edges weights form is dict{(nid1, nid2): edge_weight}
    """
    nodew = nx.get_node_attributes(graph, 'weight')
    edw = nx.get_edge_attributes(graph, 'weight')
    edgew = edw.copy()
    for key, val in edw.items():
        edgew[(key[1], key[0])] = val

    return nodew, edgew


def generate_graph_data(node_num, edge_num, weight_range=10):
    """
    generate a simple graph‘s weights of nodes and edges with
    node number is node_num, edge number is edge_num
    Args:
        node_num (int): node number in graph
        edge_num (int): edge number in graph
        weight_range (int): weight range of nodes and edges
    Return:
        nodes(set of tuple(nid, node_weight)), edges(set of tuple(nid1, nid2, edge_weight))
    """
    if weight_range is None:
        weight_range = 10

    nodes = set()
    for i in range(node_num):
        ndw = np.random.choice(range(weight_range))
        nodes |= {(i, ndw)}

    edges = set()
    cnt = edge_num
    max_edges = node_num * (node_num - 1) / 2
    if cnt > max_edges:
        cnt = max_edges
    while cnt > 0:
        u = np.random.randint(node_num)
        v = np.random.randint(node_num)
        if u == v:  # without self loop
            continue
        flg = 0
        for e in edges:  # without duplicated edges
            if set(e[:2]) == set([v, u]):
                flg = 1
                break
        if flg == 1:
            continue
        edw = np.random.choice(range(weight_range))
        edges |= {(u, v, edw)}
        cnt -= 1
    return nodes, edges


def generate_weighted_graph(nodes, edges, weight_range=10):
    """
    generate graph from nodes list and edges list which identify the nodes and edges
    that should be add in graph, and the random weight range of every node and edge.
    Args:
        nodes (list/set): list of node idex / node-weight map, element form is tuple(nid, weight)
        edges (list/set): list of edge: (e_idex1, e_idex2) / edge-weight map, element form is tuple(nid1, nid2, edge_weight)
        weight_range (int): random weight range of every node and edge
    Returns:
        g (nx.Graph): graph generated by args
    """
    g = nx.Graph()
    if isinstance(nodes, list) and isinstance(edges, list):
        for v in nodes:
            w = np.random.choice(range(weight_range))
            g.add_node(v, weight=w)

        for e in edges:
            w = np.random.choice(range(weight_range))
            g.add_edge(e[0], e[1], weight=w)
    else:
        for item in nodes:
            g.add_node(item[0], weight=item[1])

        for item in edges:
            g.add_edge(item[0], item[1], weight=item[2])
    return g


import itertools
import os
import time
from collections import defaultdict, Callable
import numpy as np
import matplotlib.pyplot as plt
from multiprocessing import Pool, cpu_count

from qiskit import QuantumCircuit, QuantumRegister, ClassicalRegister, Aer, assemble, BasicAer, transpile
from qiskit.aqua import QuantumInstance, aqua_globals
from qiskit.aqua.operators import PauliExpectation, CircuitSampler, StateFn, CircuitOp, CircuitStateFn, \
    MatrixExpectation, X, Y, Z, I

from Qcover.utils import get_graph_weights
from Qcover.backends import Backend
import warnings
warnings.filterwarnings("ignore")


class CircuitByQiskit(Backend):
    """generate a instance of CircuitByQiskit"""
    def __init__(self,
                 expectation_calc_method: str = "statistic", # or sample
                 research: str = "QAOA",
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByQiskit"""
        super(CircuitByQiskit, self).__init__()
        self._p = None
        self._origin_graph = None
        self._is_parallel = False if is_parallel is None else is_parallel
        self._expectation_calc_method = expectation_calc_method
        self._research = research

        self._nodes_weight = None
        self._edges_weight = None
        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []
        self._element_expectation = dict()

    @property
    def element_expectation(self):
        return self._element_expectation

    def get_QAOA_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ_qreg = QuantumRegister(len(graph), 'q')
        circ = QuantumCircuit(circ_qreg)

        for k in range(p):
            for nd in graph.nodes:
                u = node_to_qubit[nd]
                if k == 0:
                    circ.h(u)
                circ.rz(2 * gamma_list[k] * self._nodes_weight[nd], u)

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]], u, v)

            for nd in graph.nodes:
                circ.rx(2 * beta_list[k], node_to_qubit[nd])

        circ.measure_all()
        return circ

    def get_GHZ_circuit(self, p, graph, node_to_qubit):
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        circ_qreg = QuantumRegister(len(graph), 'q')
        circ = QuantumCircuit(circ_qreg)

        pivot = len(self._origin_graph) // 2

        for k in range(p):
            for nd in graph.nodes:
                if k == 0:
                    circ.h(node_to_qubit[nd])

            for edge in graph.edges:
                    u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                    if u == v:
                        continue
                    circ.rzz(2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]], u, v)  #

            for nd in graph.nodes:
                if nd != pivot:
                    circ.rx(2 * beta_list[k], node_to_qubit[nd])

            circ.measure_all()
            return circ

    def create_qaoa_cir(self, G):
        N_qubits = len(G.nodes())
        p = len(self._pargs) // 2
        qc = QuantumCircuit(N_qubits)

        gamma = self._pargs[:p]
        beta = self._pargs[p:]

        for i in range((N_qubits - 1) // 2):
            qc.h(i)

        qc.ry(np.pi/2, (N_qubits - 1) // 2)

        for i in range((N_qubits + 1) // 2, N_qubits):
            qc.h(i)

        for irep in range(p):
            for pair in list(G.edges()):
                qc.rzz(-2 * gamma[irep], pair[0], pair[1])

            # for i in range(N_qubits):
            #   qc.rx(2*beta[irep], i)

            for i in range((N_qubits - 1) // 2):
                qc.rx(2 * beta[irep], i)
            for i in range((N_qubits + 1) // 2, N_qubits):
                qc.rx(2 * beta[irep], i)

        qc.measure_all()

        return qc

    def Ising_obj(self, x, G):
        obj = 0
        for i, j in G.edges():
            if x[i] == x[j]:
                obj -= 1
            else:
                obj += 1
        return obj

    def compute_expectation(self, counts, G):
        avg = 0
        sum_count = 0
        for bitstring, count in counts.items():
            obj = self.Ising_obj(bitstring, G)
            avg += obj * count
            sum_count += count
        return avg / sum_count

    def get_expectation(self, G, p, shots=1024):
        p = self._p if p is None else p
        backend = Aer.get_backend('qasm_simulator')
        backend.shots = shots

        qc = self.create_qaoa_cir(G)
        counts = backend.run(qc, seed_simulator=10, nshots=1024 * 100).result().get_counts()

        avg = 0
        sum_count = 0
        for bitstring, count in counts.items():
            obj = self.Ising_obj(bitstring, G)
            avg += obj * count
            sum_count += count
        return avg / sum_count

    # def get_expectation(self, element_graph, p=None):
    #     if self._is_parallel is False:
    #         p = self._p if p is None else p
    #         original_e, graph = element_graph
    #     else:
    #         p = self._p if len(element_graph) == 1 else element_graph[1]
    #         original_e, graph = element_graph[0]
    #
    #     node_to_qubit = defaultdict(int)
    #     node_list = list(graph.nodes)
    #     for i in range(len(node_list)):
    #         node_to_qubit[node_list[i]] = i
    #
    #     if self._research == "QAOA":
    #         circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
    #     elif self._research == "GHZ":
    #         pivot = len(self._origin_graph) // 2
    #         if pivot in graph:
    #             circ = self.get_GHZ_circuit(p, graph, node_to_qubit)
    #         else:
    #             circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
    #     # circ.draw(output='mpl', interactive=True)
    #     if isinstance(original_e, int):
    #         weight = self._nodes_weight[original_e]
    #     else:
    #         weight = self._edges_weight[original_e]
    #
    #     backend = Aer.get_backend('qasm_simulator')
    #     counts = backend.run(circ, seed_simulator=43, nshots=1024).result().get_counts()
    #
    #     def Ising_obj(x, G):
    #         obj = 0
    #         for ed in G.edges():
    #             if x[node_to_qubit[ed[0]]] == x[node_to_qubit[ed[1]]]:
    #                 obj -= self._edges_weight[ed]
    #             else:
    #                 obj += self._edges_weight[ed]
    #         return obj
    #
    #     avg = 0
    #     sum_count = 0
    #     for bitstring, count in counts.items():
    #         obj = Ising_obj(bitstring, graph)
    #         avg += obj * count
    #         sum_count += count
    #     exp_res = avg / sum_count
    #
    #     return weight, exp_res

    def expectation_calculation(self, p=None):
        if self._nodes_weight is None or self._edges_weight is None:
            nodes_weight, edges_weight = get_graph_weights(self._origin_graph)
            self._nodes_weight, self._edges_weight = nodes_weight, edges_weight

        self._element_expectation = {}
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        cpu_num = cpu_count()
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        res = 0
        # st = time.time()
        for item in self._element_to_graph.items():
            original_e, G = item
            w_i = self._nodes_weight[original_e]
            exp_i = self.get_expectation(G, p)
            # if isinstance(item[0], tuple):  #origin format of RQAOA
            self._element_expectation[item[0]] = exp_i
            res += w_i * exp_i  #
        # ed = time.time()
        # print("exp one cost:", ed - st)
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        args = list(itertools.product(self._element_to_graph.items(), [p]))
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, args))
        # circ_res.append(pool.map(self.get_expectation, list(self._element_to_graph.items()), chunksize=1))

        pool.terminate()  # pool.close()
        pool.join()
        res = 0
        for it in circ_res[0]:
            res += it[0] * it[1]

        # res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def optimization_visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="Qiskit")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()

    # below is functions not used now
    def graph_to_circuit(self, params, graph, p=None, original_e=None):
        """
        transform the graph to circuit according to the computing_framework
        Args:
            graph (nx.Graph): graph to be transformed to circuit
            params (List): Optimal parameters   #np.array
            original_e (Optional[None, int, tuple])
        Return:
            if original_e=None, then the graph is the whole original graph generated by
            generate_weighted_graph(), so just return the circuit transformed by it

            if original_e is a int, then the subgraph is generated by node(idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[], and the circuit

            if original_e is a tuple, then the subgraph is generated by edge(node idx = original_e
            in whole graph), so return the it's idx mapped by node_to_qubit[] as
            tuple(mapped node_id1, mapped node_id2), and the circuit
        """

        # generate node_to_qubit[] map from node id in graph to [0, node_num in graph - 1]
        # so the subgraph would has it's own qubits number
        p = self._p if p is None else p

        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        self._pargs = params
        if self._research == "QAOA":
            circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
        elif self._research == "GHZ":
            circ = self.get_GHZ_circuit(p, graph, node_to_qubit)

        if original_e is None:
            return original_e, circ
        elif isinstance(original_e, int):
            return node_to_qubit[original_e], circ
        else:
            return (node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]), circ

    def get_result_statevector(self, params, graph):

        params = params.flatten()
        _, circ = self.graph_to_circuit(params, graph)
        circ.save_statevector()
        # circ.draw(output='mpl', interactive=True)

        sim = Aer.get_backend('aer_simulator')
        result = sim.run(circ).result()
        out_state = result.get_statevector()

        return out_state

    def get_result_counts(self, params):  # 直接将最优参数值带入整个大电路中运行qubit将只能到18

        params = params.flatten()
        _, circ = self.graph_to_circuit(params, self._origin_graph)
        circ.draw(output='mpl', interactive=True)
        # circ.measure_all()   # if use sample method, can't use measure

        sim = Aer.get_backend('qasm_simulator')
        transpiled_qc = transpile(circ, sim)
        job_sim = sim.run(transpiled_qc, shots=1024)
        hist = job_sim.result().get_counts(transpiled_qc)
        # for key, val in hist.items():
        #     print("%s has %s" % (key, str(val)))

        # from qiskit.visualization import plot_histogram
        # plot_histogram(hist)
        # plt.show()
        return hist

    def sampling_visualization(self, counts):
        from qiskit.visualization import plot_histogram
        plot_histogram(counts)
        plt.show()



import sys
import numpy as np
from typing import Iterator, Optional, Union, Callable, Dict, Any
from Qcover.exceptions import ArrayShapeError

class GradientDescent:

    def __init__(self,
                 options: dict = None,  # {'maxiter':300, 'learning_rate':0.0005, 'tol':1e-6, 'perturbation':0.1},
                 initial_point: Optional[np.ndarray] = None) -> None:

        super().__init__()

        self._p = None
        self._options = options
        self._initial_point = initial_point

    def _minimize(self, loss):
        if 'learning_rate' in self._options:
            lr = self._options["learning_rate"]
        else:
            lr = 0.01

        if 'tol' in self._options:
            tol = self._options['tol']
        else:
            tol = 1e-6

        if 'maxiter' in self._options:
            maxiter = self._options['maxiter']
        else:
            maxiter = 500

        eps = 0.01 if "perturbation" not in self._options else self._options['perturbation']

        # prepare some initials
        x = np.asarray(self._initial_point)
        nfevs = 0

        for _ in range(1, maxiter + 1):
            delta = np.random.random() - 0.5
            # delta = np.random.randint(0, 1, size=len(self._initial_point)) - 0.5
            diff = loss(x + delta) - loss(x - delta)  #
            grad = 0.5 * diff / delta  #

            x_next = x - lr * grad
            stepsize = np.linalg.norm(grad)  #the distance between grad to zero
            x = x_next
            nfevs += 1

            # check termination
            if stepsize < tol:
                break

        return x, loss(x), nfevs

    def optimize(self, objective_function):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * self._p)])
        else:
            try:
                if len(self._initial_point) != 2 * self._p:
                    raise ArrayShapeError("The shape of initial parameters is not match with p")
            except ArrayShapeError as e:
                print(e)
                sys.exit()

        return self._minimize(objective_function)






import os
import time
import itertools

from collections import defaultdict, Callable
import numpy as np
import sympy
import matplotlib.pyplot as plt
from multiprocessing import Pool
import networkx as nx

from Qcover.simulator import Qcircuit, Qcodes
from Qcover.backends import Backend
from Qcover.utils import get_graph_weights
import warnings
warnings.filterwarnings("ignore")


class CircuitByQton(Backend):
    """generate a instance of CircuitByQton"""

    def __init__(self,
                 research: str = "QAOA",
                 expectation_calc_method: str = "statevector",
                 is_parallel: bool = None) -> None:
        """initialize a instance of CircuitByCirq"""
        super(CircuitByQton, self).__init__()

        self._p = None
        self._origin_graph = None
        self._is_parallel = False if is_parallel is None else is_parallel
        self._expectation_calc_method = expectation_calc_method
        self._research = research

        self._nodes_weight = None
        self._edges_weight = None
        self._element_to_graph = None
        self._pargs = None
        self._expectation_path = []
        self._element_expectation = dict()

    @property
    def element_expectation(self):
        return self._element_expectation

    @staticmethod
    def get_operator(self):
        pass

    def get_QAOA_circuit(self, p, graph, node_to_qubit):
        circ = Qcircuit(len(graph))
        circ.mode = self._expectation_calc_method
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]
        for k in range(p):
            for i in graph.nodes:
                u = node_to_qubit[i]
                if k == 0:
                    circ.h(u)
                circ.rz(u, 2 * gamma_list[k] * self._nodes_weight[i])

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.rzz(u, v, 2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]])

            for nd in graph.nodes:
                u = node_to_qubit[nd]
                circ.rx(u, 2 * beta_list[k])

        return circ

    def get_GHZ_circuit(self, p, graph, node_to_qubit):
        circ = Qcircuit(len(graph))
        circ.mode = self._expectation_calc_method
        gamma_list, beta_list = self._pargs[: p], self._pargs[p:]

        pivot = len(self._origin_graph) // 2
        for k in range(p):
            for nd in graph.nodes:
                u = node_to_qubit[nd]
                if k == 0:
                    circ.h(u)

            for edge in graph.edges:
                u, v = node_to_qubit[edge[0]], node_to_qubit[edge[1]]
                if u == v:
                    continue
                circ.rzz(u, v, 2 * gamma_list[k] * self._edges_weight[edge[0], edge[1]])

            for nd in graph.nodes:
                if nd != pivot:
                    u = node_to_qubit[nd]
                    circ.rx(u, 2 * beta_list[k])

        return circ

    def get_expectation(self, element_graph, p=None):  #,
        """
        calculate the expectation of the subgraph
        Args:
            element_graph: tuple of (original node/edge, subgraph)

        Returns:
            expectation of the subgraph
        """

        if self._is_parallel is False:
            p = self._p if p is None else p
            original_e, graph = element_graph
        else:
            p = self._p if len(element_graph) == 1 else element_graph[1]
            original_e, graph = element_graph[0]

        node_to_qubit = defaultdict(int)
        node_list = list(graph.nodes)
        for i in range(len(node_list)):
            node_to_qubit[node_list[i]] = i

        if self._research == "QAOA":
            circ = self.get_QAOA_circuit(p, graph, node_to_qubit)
        elif self._research == "GHZ":
            pivot = len(self._origin_graph) // 2
            if pivot in graph:
                circ = self.get_GHZ_circuit(p, graph, node_to_qubit)
            else:
                circ = self.get_QAOA_circuit(p, graph, node_to_qubit)

        state_circ = circ.state.copy()
        if isinstance(original_e, int):
            weight = self._nodes_weight[original_e]
            circ.z(node_to_qubit[original_e])
            # op = self.get_operator(node_to_qubit[original_e], len(node_list))
        else:
            weight = self._edges_weight[original_e]
            circ.z(node_to_qubit[original_e[0]])
            circ.z(node_to_qubit[original_e[1]])
            # op = self.get_operator((node_to_qubit[original_e[0]], node_to_qubit[original_e[1]]), len(node_list))

        state_opc = circ.state.copy()
        exp_res = np.inner(state_opc.conj(), state_circ)

        return weight, exp_res

    def expectation_calculation(self, p=None):
        if self._nodes_weight is None or self._edges_weight is None:
            nodes_weight, edges_weight = get_graph_weights(self._origin_graph)
            self._nodes_weight, self._edges_weight = nodes_weight, edges_weight

        self._element_expectation = {}
        if self._is_parallel:
            return self.expectation_calculation_parallel(p)
        else:
            return self.expectation_calculation_serial(p)

    def expectation_calculation_serial(self, p=None):
        res = 0
        for item in self._element_to_graph.items():
            w_i, exp_i = self.get_expectation(item, p)
            if isinstance(item[0], tuple):
                self._element_expectation[item[0]] = exp_i
            res += w_i * exp_i

        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def expectation_calculation_parallel(self, p=None):
        cpu_num = 1
        os.environ['OMP_NUM_THREADS'] = str(cpu_num)
        os.environ['OPENBLAS_NUM_THREADS'] = str(cpu_num)
        os.environ['MKL_NUM_THREADS'] = str(cpu_num)
        os.environ['VECLIB_MAXIMUM_THREADS'] = str(cpu_num)
        os.environ['NUMEXPR_NUM_THREADS'] = str(cpu_num)

        circ_res = []
        args = list(itertools.product(self._element_to_graph.items(), [p]))
        pool = Pool(os.cpu_count())
        circ_res.append(pool.map(self.get_expectation, args))
        # circ_res.append(pool.map(self.get_expectation, list(self._element_to_graph.items()), chunksize=1))

        pool.terminate()  # pool.close()
        pool.join()
        res = 0
        for it in circ_res[0]:
            res += it[0] * it[1]
        # res = sum(circ_res[0])
        print("Total expectation of original graph is: ", res)
        self._expectation_path.append(res)
        return res

    def optimization_visualization(self):
        plt.figure()
        plt.plot(range(1, len(self._expectation_path) + 1), self._expectation_path, "ob-", label="qulacs")
        plt.ylabel('Expectation value')
        plt.xlabel('Number of iterations')
        plt.legend()
        plt.show()

    # def get_result_counts(self, params, graph):  # 直接将最优参数值带入整个大电路中运行qubit将只能到18
    #
    #     params = params.flatten()
    #     _, circ = self.graph_to_circuit(params, graph)
    #     circ.draw(output='mpl', interactive=True)
    #     circ.measure_all()   # if use sample method, can't use measure
    #
    #     sim = Aer.get_backend('qasm_simulator')
    #     transpiled_qc = transpile(circ, sim)
    #     job_sim = sim.run(transpiled_qc, shots=1024)
    #     hist = job_sim.result().get_counts(transpiled_qc)
    #     for key, val in hist.items():
    #         print("%s has %s" % (key, str(val)))
    #
    #     from qiskit.visualization import plot_histogram
    #     plot_histogram(hist)
    #     plt.show()
    #     return hist



import sys
from typing import Optional
import logging
import numpy as np
from scipy import optimize as opt
from Qcover.optimizers import Optimizer
from Qcover.exceptions import ArrayShapeError

logger = logging.getLogger(__name__)


class SLSQP(Optimizer):
    """
    SLSQP: a numerical optimization method for constrained problems

    based on scipy.optimize.minimize SLSQP.
    For further detail, please refer to
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.minimize.html
    """

    # pylint: disable=unused-argument
    def __init__(self,
                 options: dict = None,
                 initial_point: Optional[np.ndarray] = None) -> None:
        """
        Args:
            options: some optional setting parameters such as:
                maxiter: Maximum number of function evaluations.
                disp: Set to True to print convergence messages.
                rhobeg: Reasonable initial changes to the variables.
                tol: Final accuracy in the optimization (not precisely guaranteed).
                     This is a lower bound on the size of the trust region.
        """
        super().__init__()
        self._p = None
        self._options = options
        self._initial_point = initial_point

    def optimize(self, objective_function):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * self._p)])
        else:
            try:
                if len(self._initial_point) != 2 * self._p:
                    raise ArrayShapeError("The shape of initial parameters is not match with p")
            except ArrayShapeError as e:
                print(e)
                sys.exit()

        res = opt.minimize(objective_function,
                           x0=np.array(self._initial_point),
                           args=self._p,
                           method='SLSQP',
                           jac=None,
                           options=self._options
                           )

        return res.x, res.fun, res.nfev

# opts = SLSQP(options={'maxiter': 100,
#         'ftol': 1e-06,
#         'iprint': 1,
#         'disp': False,
#         'eps': 1.4901161193847656e-08,
#         'finite_diff_rel_step': None})


from typing import Optional
import logging
import numpy as np

from scipy import optimize as opt

logger = logging.getLogger(__name__)

class Optimizer:
    def __init__(self, *args):
        pass

    def optimize(self, objective_function):
        pass


"""Backend interface"""

from enum import IntEnum
import logging
from abc import ABC, abstractmethod
import numpy as np

logger = logging.getLogger(__name__)

# pylint: disable=invalid-name

class Backend(ABC):
    """Base class for backend."""

    # @abstractmethod
    # def __init__(self, *args):
    #     pass

    @abstractmethod
    def __init__(self,
                 p: int = 1,
                 nodes_weight: list = None,
                 edges_weight: list = None,
                 is_parallel: bool = None) -> None:
        self._p = p
        self._nodes_weight = nodes_weight
        self._edges_weight = edges_weight
        self._is_parallel = False if is_parallel is None else is_parallel

        self._element_to_graph = None
        self._pargs = None
        self._element_expectation = dict()

    # @abstractmethod
    # def get_operator(self, *args):
    #     pass

    @abstractmethod
    def get_expectation(self, *args):
        pass

    @abstractmethod
    def expectation_calculation(self):
        pass

    @abstractmethod
    def get_result_counts(self, params):
        pass

    @abstractmethod
    def optimization_visualization(self):
        pass

    @abstractmethod
    def sampling_visualization(self, counts):
        pass


import sys
from typing import Optional
import logging
import numpy as np
from scipy import optimize as opt
from Qcover.optimizers import Optimizer
from Qcover.exceptions import ArrayShapeError
import warnings
warnings.filterwarnings("ignore")

logger = logging.getLogger(__name__)

class SPSA(Optimizer):
    # pylint: disable=unused-argument
    def __init__(self,
                 options: dict = None,  # {'maxiter':300, 'disp':True, 'rhobeg': 1.0, 'tol':1e-6},
                 initial_point: Optional[np.ndarray] = None) -> None:
        """
        Args:
            options: some optional setting parameters such as:
                maxiter: Maximum number of function evaluations.
                disp: Set to True to print convergence messages.
                rhobeg: Reasonable initial changes to the variables.
                tol: Final accuracy in the optimization (not precisely guaranteed).
                     This is a lower bound on the size of the trust region.
        """
        super().__init__()
        self._p = None
        self._options = options
        self._initial_point = initial_point

    def _minimize(self, loss):
        if 'A' in self._options:
            A = self._options["A"]
        else:
            A = 20

        if 'R' in self._options:
            R = self._options['R']
        else:
            R = 0.2

        if 'a0' in self._options:
            a0 = self._options["a0"]
        else:
            a0 = 1.3

        if 'c0' in self._options:
            c0 = self._options['c0']
        else:
            c0 = 0.2

        if 'tol' in self._options:
            tol = self._options['tol']
        else:
            tol = 1e-6

        if 'maxiter' in self._options:
            maxiter = self._options['maxiter']
        else:
            maxiter = 500

        # a0 = 1.2
        # c0 = 1.0
        # prepare some initials
        x = np.asarray(self._initial_point)
        nfevs = 0

        for i in range(1, maxiter + 1):
            a = a0 / (i + 1) ** A
            c = c0 / (i + 1) ** R
            # delta = np.random.random() - 0.5
            delta = (2 * np.random.randint(0, 2, size=len(self._initial_point)) - 1) * c
            diff = loss(x + delta) - loss(x - delta)  #
            grad = 0.5 * diff / delta  #

            x_next = x - a * grad
            x = x_next
            stepsize = np.linalg.norm(grad)  # the distance between grad to zero
            nfevs += 1
            # check termination
            if stepsize < tol:
                break

        return x, loss(x), nfevs

    def optimize(self, objective_function):
        if self._initial_point is None:
            self._initial_point = np.array([np.random.random() for x in range(2 * self._p)])
        else:
            try:
                if len(self._initial_point) != 2 * self._p:
                    raise ArrayShapeError("The shape of initial parameters is not match with p")
            except ArrayShapeError as e:
                print(e)
                sys.exit()

        return self._minimize(objective_function)